import argparse
import pytest

from semeio.jobs.rft.zonemap import ZoneMap


def test_zonemap():
    zones_at_k_value = {1: ["Zone1", "Zone2"], 3: ["Zone3"]}
    zone_map = ZoneMap(zones_at_k_value)
    assert zone_map.has_relationship("Zone1", 1)
    assert zone_map.has_relationship("Zone2", 1)
    assert not zone_map.has_relationship("Zone1", 3)
    assert zone_map.has_relationship("Zone3", 3)


@pytest.fixture()
def initdir(tmpdir):
    tmpdir.chdir()
    valid_data = """
1 zone1 zone3
2 zone2
3 zone2
"""
    tmpdir.join("valid_zonemap.txt").write(valid_data)
    comment = "-- this is a comment"
    valid_data = comment + "\n" + valid_data + comment
    tmpdir.join("valid_zonemap_with_comments.txt").write(valid_data)

    valid_weird_data = """
1 zone1 zone3 2 3
2 zone2
3 zone2
"""
    tmpdir.join("valid_weird_zonemap.txt").write(valid_weird_data)

    invalid_data = """
0 zone1
2 zone2
3 zone2
"""
    tmpdir.join("invalid_starts_at_0.txt").write(invalid_data)

    invalid_data = """
zone1 2
2 zone2
3 zone2
"""
    tmpdir.join("invalid_format_string_first.txt").write(invalid_data)

    invalid_data = """
1
2 zone2
3 zone2
"""
    tmpdir.join("invalid_format_missing_zone.txt").write(invalid_data)


def test_load_data(initdir):

    for fname in [
        "valid_zonemap.txt",
        "valid_zonemap_with_comments.txt",
        "valid_weird_zonemap.txt",
    ]:
        zone_map = ZoneMap.load_and_parse_zonemap_file(fname)
        assert zone_map.has_relationship("zone1", 0)
        assert zone_map.has_relationship("zone3", 0)
        assert zone_map.has_relationship("zone2", 2)
        assert zone_map.has_relationship("zone2", 1)
        assert not zone_map.has_relationship("zone1", 1)
        assert not zone_map.has_relationship("zone1", 2)


def test_invalid_load(initdir):
    errors = [
        "k values cannot be 0, must start at 1",
        "k must be integer, was zone1",
        "Number of zonenames must be 1 or more",
        "ZoneMap file non_existing not found",
    ]

    fnames = [
        "invalid_starts_at_0.txt",
        "invalid_format_string_first.txt",
        "invalid_format_missing_zone.txt",
        "non_existing",
    ]

    for fname, error in zip(fnames, errors):
        with pytest.raises(argparse.ArgumentTypeError) as msgcontext:
            ZoneMap.load_and_parse_zonemap_file(fname)
        assert error in msgcontext.value.args[0]



# pylint: disable=missing-docstring

import os
import semeio.hook_implementations.jobs
from semeio.jobs.scripts import (
    csv_export2,
)
from semeio.workflows.misfit_preprocessor import misfit_preprocessor
from semeio.workflows.spearman_correlation_job import spearman_correlation
from semeio.workflows.correlated_observations_scaling import cos
from ert_shared.plugins.plugin_manager import ErtPluginManager


def test_hook_implementations():
    pm = ErtPluginManager(
        plugins=[
            semeio.hook_implementations.jobs,
            spearman_correlation,
            misfit_preprocessor,
            cos,
            csv_export2,
        ]
    )

    expected_jobs = {
        "DESIGN_KW": "semeio/jobs/config_jobs/DESIGN_KW",
        "DESIGN2PARAMS": "semeio/jobs/config_jobs/DESIGN2PARAMS",
        "STEA": "semeio/jobs/config_jobs/STEA",
        "GENDATA_RFT": "semeio/jobs/config_jobs/GENDATA_RFT",
        "PYSCAL": "semeio/jobs/config_jobs/PYSCAL",
        "INSERT_NOSIM": "semeio/jobs/config_jobs/INSERT_NOSIM",
        "REMOVE_NOSIM": "semeio/jobs/config_jobs/REMOVE_NOSIM",
    }
    installable_jobs = pm.get_installable_jobs()
    for wf_name, wf_location in expected_jobs.items():
        assert wf_name in installable_jobs
        assert installable_jobs[wf_name].endswith(wf_location)

    assert set(installable_jobs.keys()) == set(expected_jobs.keys())

    expected_workflow_jobs = [
        "CORRELATED_OBSERVATIONS_SCALING",
        "SPEARMAN_CORRELATION",
        "CSV_EXPORT2",
        "MISFIT_PREPROCESSOR",
    ]
    installable_workflow_jobs = pm.get_installable_workflow_jobs()
    for wf_name, wf_location in installable_workflow_jobs.items():
        assert wf_name in expected_workflow_jobs
        assert os.path.isfile(wf_location)

    assert set(installable_workflow_jobs.keys()) == set(expected_workflow_jobs)


def test_hook_implementations_job_docs():
    pm = ErtPluginManager(plugins=[semeio.hook_implementations.jobs])

    installable_jobs = pm.get_installable_jobs()

    docs = pm.get_documentation_for_jobs()

    assert set(docs.keys()) == set(installable_jobs.keys())

    for job_name in installable_jobs.keys():
        assert docs[job_name]["description"] != ""
        assert docs[job_name]["category"] != "other"



import itertools
import argparse

import pandas as pd

import pytest

from semeio.jobs.rft.trajectory import TrajectoryPoint, Trajectory
from semeio.jobs.rft.zonemap import ZoneMap


@pytest.mark.parametrize(
    "line, expected_zone", [("0 1 2.2 3", None), ("0 1 2.2 3 zone", "zone")]
)
def test_load_from_line(line, expected_zone):
    point = TrajectoryPoint(*Trajectory.parse_trajectory_line(line))
    assert point.utm_x == 0
    assert point.utm_y == 1
    assert point.measured_depth == 2.2
    assert point.true_vertical_depth == 3
    assert point.zone == expected_zone


@pytest.mark.parametrize(
    "line, expected_error",
    [
        ("1", "Trajectory data file not on correct format"),
        ("a b c", "Trajectory data file not on correct format"),
        ("a b c d e", "Error: Failed to extract data from line"),
    ],
)
def test_invalid_load_from_line(line, expected_error):
    with pytest.raises(argparse.ArgumentTypeError) as err_context:
        TrajectoryPoint(*Trajectory.parse_trajectory_line(line))
    assert expected_error in err_context.value.args[0]


@pytest.mark.parametrize(
    "k_value, expected_validation", [(1, True), (3, False), (4, False)]
)
def test_validate_zone(k_value, expected_validation):
    zones_at_k_value = {1: ["Zone1", "Zone2"], 3: ["Zone3"]}
    zone_map = ZoneMap(zones_at_k_value)

    point = TrajectoryPoint(0, 1, 2, 3, "Zone1")
    point.set_ijk([0, 0, k_value])

    assert not point.valid_zone
    point.validate_zone(zone_map)

    assert point.valid_zone == expected_validation


def test_validate_non_existing_zone():
    zones_at_k_value = {1: ["Zone1", "Zone2"], 3: ["Zone3"]}
    zone_map = ZoneMap(zones_at_k_value)

    point = TrajectoryPoint(0, 1, 2, 3, "non existing")
    point.set_ijk((0, 0, 1))
    point.validate_zone(zone_map)
    assert not point.valid_zone


ijks = [None, (0, 0, 0)]
pressures = [None, 1]
validations = [False, True]


@pytest.mark.parametrize(
    "ijk, pressure, validation", itertools.product(ijks, pressures, validations)
)
def test_is_active(ijk, pressure, validation):

    point = TrajectoryPoint(0, 1, 2, 3, "Zone1")
    point.set_ijk(ijk)
    point.pressure = pressure
    point.valid_zone = validation
    assert all([ijk, pressure, validation]) == point.is_active()

    assert 1 if all([ijk, pressure, validation]) else -1 == point.get_pressure()


@pytest.fixture()
def initdir(tmpdir):
    tmpdir.chdir()
    valid_data = """
0 1 2.2 3
4 5 6 7 zone
"""
    # (utm_x utm_y md tvd zone_name)
    tmpdir.join("valid_trajectories.txt").write(valid_data)
    comment = "-- this is a comment"
    valid_data = comment + "\n" + valid_data + comment
    tmpdir.join("valid_trajectories_with_comments.txt").write(valid_data)


@pytest.mark.parametrize(
    "fname", ["valid_trajectories.txt", "valid_trajectories_with_comments.txt"]
)
def test_load(initdir, fname):
    expected_utmxs = [0, 4]
    trajectory = Trajectory.load_from_file(fname)

    for expected_utmx, trajectorypoint in zip(
        expected_utmxs, trajectory.trajectory_points
    ):
        assert trajectorypoint.utm_x == expected_utmx


@pytest.mark.parametrize(
    "fname", ["valid_trajectories.txt", "valid_trajectories_with_comments.txt"]
)
def test_dframe_trajectory(initdir, fname):
    """Test dataframe representation of a trajectory not having
    any attached Eclipse simulation results"""
    dframe = Trajectory.load_from_file(fname).to_dataframe()

    assert isinstance(dframe, pd.DataFrame)

    # Dataframe lengths should be the same as number of non-empty lines
    # in txt input:
    assert len(dframe) == len(
        [
            line
            for line in open(fname).readlines()
            if line.strip() and not line.strip().startswith("--")
        ],
    )

    # Check that we have the input columns which defines the trajectory:
    assert {"utm_x", "utm_y", "measured_depth", "true_vertical_depth", "zone"}.issubset(
        set(dframe.columns)
    )

    # grid_ijk is a temp column, never to be present in output
    assert "grid_ijk" not in dframe
    # and since there is no Eclipse results attached, we can't have these either:
    assert "i" not in dframe
    assert "j" not in dframe
    assert "k" not in dframe

    # Check casing for column names, ensuring consistency in this particular
    # dataframe:
    assert list(dframe.columns) == [colname.lower() for colname in dframe.columns]

    # pressure should not be there when there is no data for it
    # (no Eclipse simulation results attached in this particular test function)
    assert "pressure" not in dframe

    # These columns should be in place, to signify there is not data for them.
    # (Eclipse simulation results would be needed for any of these to be True)
    assert set(dframe["valid_zone"]) == {False}
    assert set(dframe["is_active"]) == {False}

    # Check dataframe sorting:
    assert (
        dframe.sort_values("measured_depth")["measured_depth"]
        == dframe["measured_depth"]
    ).all()


@pytest.mark.parametrize(
    "dframe, tuplecolumn, components",
    [
        (pd.DataFrame(columns=["grid_ik"], data=[[(1, 2)]]), "grid_ik", ["I", "J"]),
        (pd.DataFrame(columns=["grid_i"], data=[[(1,)]]), "grid_i", ["I"]),
        (
            pd.DataFrame(columns=["grid_i"], data=[[[1, 2, 3]]]),
            "grid_i",
            ["I", "J", "K"],
        ),
        (
            pd.DataFrame(columns=["grid_i"], data=[[(1, 2)], [(3, 4)]]),
            "grid_i",
            ["I", "J"],
        ),
        (
            pd.DataFrame(columns=["grid_i"], data=[[(1, 2)], [(3, 4)]]),
            "grid_i",
            ["I", "J"],
        ),
        (
            pd.DataFrame(
                columns=["grid_i", "extra"], data=[[(1, 2), "foo"], [(3, 4), "bar"]]
            ),
            "grid_i",
            ["I", "J"],
        ),
    ],
)
def test_tuple_column_splitter(dframe, tuplecolumn, components):
    splitdf = Trajectory.split_tuple_column(
        dframe, tuplecolumn=tuplecolumn, components=components
    )

    assert tuplecolumn in dframe  # Ensure we have not touched the input
    assert tuplecolumn not in splitdf
    assert len(dframe) == len(splitdf)
    assert {val for tup in dframe[tuplecolumn] for val in tup} == {
        val for tup in splitdf[components].values for val in tup
    }
    for comp in components:
        assert comp in splitdf
    assert len(splitdf.columns) == len(dframe.columns) - 1 + len(components)


@pytest.mark.parametrize(
    "raises, dframe, tuplecolumn, components",
    [
        (
            ValueError,
            pd.DataFrame(columns=["grid_ik"], data=[[(1, 2)]]),
            "grid_ik",
            ["I", "J", "K"],
        ),
        (
            ValueError,
            pd.DataFrame(columns=["grid_ik"], data=[[(1, 2)], [(1, 2, 3)]]),
            "grid_ik",
            ["I", "J", "K"],
        ),
        (
            ValueError,
            pd.DataFrame(columns=["grid_ik"], data=[[(1, 2)], [(1, 2, 3)]]),
            "grid_ik",
            ["I", "J"],
        ),
    ],
)
def test_tuple_column_splitter_errors(raises, dframe, tuplecolumn, components):
    with pytest.raises(raises):
        Trajectory.split_tuple_column(
            dframe, tuplecolumn=tuplecolumn, components=components
        )


def test_tuple_column_splitter_explicit():
    # Checks that None-rows are handled, and default parameters for split_tuple_column()
    dframe = Trajectory.split_tuple_column(
        pd.DataFrame(columns=["grid_ijk"], data=[[None], [(1, 2, 3)]])
    )
    assert len(dframe) == 2
    assert {"i", "j", "k"}.issubset(set(dframe.columns))


def test_non_existing_file():

    with pytest.raises(argparse.ArgumentTypeError) as err_context:
        Trajectory.load_from_file("non_existing")

    assert (
        "Warning: Trajectory file non_existing not found!" == err_context.value.args[0]
    )



from copy import deepcopy

import pytest
import configsuite

from semeio.workflows.correlated_observations_scaling.job_config import (
    _expand_input,
    _min_value,
    _to_int_list,
    build_schema,
)


@pytest.mark.parametrize(
    "valid_input",
    [
        "0,1,2-5",
        [0, 1, "2-5"],
        [0, 1, 2, 3, 4, 5],
        [0, 1, "2-3", "4-5"],
        "0-5",
        "0-1,2,3-5",
        ["0,1,2-5"],
    ],
)
def test_to_int_list(valid_input):
    expected_result = list(range(6))
    assert _to_int_list(valid_input) == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result", [(-1, False), (0, True), (1, True)]
)
def test_min_value(test_input, expected_result):
    assert _min_value(test_input).__bool__() == expected_result


def test_expand_input_no_modification():
    expected_result = {
        "UPDATE_KEYS": {"keys": [{"key": "key_4"}, {"key": "key_5"}, {"key": "key_6"}]},
        "CALCULATE_KEYS": {
            "keys": [{"key": "key_1"}, {"key": "key_2"}, {"key": "key_3"}],
            "threshold": 1.0,
        },
    }

    assert _expand_input(deepcopy(expected_result)) == expected_result


def test_expand_input_modification():
    keys = [{"key": "key_1"}, {"key": "key_2"}, {"key": "key_3"}]
    test_input = {"CALCULATE_KEYS": {"keys": keys, "threshold": 1.0}}

    expected_results = deepcopy(test_input)
    expected_results["UPDATE_KEYS"] = {"keys": keys}

    assert _expand_input(test_input) == expected_results


@pytest.mark.parametrize(
    "valid_config",
    [
        {
            "CALCULATE_KEYS": {
                "keys": [{"key": "first_key"}, {"key": "second_key"}],
                "alpha": 3,
                "std_cutoff": 0.0001,
            }
        },
        {
            "CALCULATE_KEYS": {
                "keys": [{"key": "first_key"}, {"key": "second_key"}],
                "alpha": 3,
                "std_cutoff": 0.0001,
            },
            "UPDATE_KEYS": {"keys": [{"index": [1, 2, 3], "key": "first_key"}]},
        },
    ],
)
def test_valid_config_setup(valid_config):

    schema = build_schema()
    config = configsuite.ConfigSuite(
        valid_config,
        schema,
        deduce_required=True,
    )
    assert config.valid


@pytest.mark.parametrize(
    "test_input,expected_errors",
    [
        pytest.param(
            {"UPDATE_KEYS": {"keys": [{"index": "1", "key": "a_key"}]}},
            [
                "keys must be provided for CALCULATE_KEYS is false on input '()'",
                "Missing key: std_cutoff",
                "Missing key: alpha",
            ],
            id="invalid_missing_required_CALCULATE_KEYS",
        ),
        pytest.param(
            {
                "CALCULATE_KEYS": {
                    "keys": [{"index": "1"}],
                    "alpha": 3,
                    "std_cutoff": 0.001,
                },
            },
            # The error msg is applied to both CALCULATE_KEYS and UPDATE_KEYS
            ["Missing key: key", "Missing key: key"],
            id="invalid_missing_required_CALCULATE_KEYS_keys",
        ),
        pytest.param(
            {
                "CALCULATE_KEYS": {
                    "keys": [{"key": "first_key"}, {"key": "second_key"}],
                    "alpha": 3,
                    "std_cutoff": 0.001,
                },
                "UPDATE_KEYS": {"keys": [{"index": [-1, 2, 3], "key": "first_key"}]},
            },
            [
                "Elements can not be negative",
                "Minimum value of index must be >= 0 is false on input '-1'",
            ],
            id="invalid_negative_index",
        ),
        pytest.param(
            {"CALCULATE_KEYS": {"keys": [{"key": "1"}], "std_cutoff": 0.001}},
            # The error msg is applied to both CALCULATE_KEYS and UPDATE_KEYS
            ["Missing key: alpha"],
            id="invalid_missing_required_CALCULATE_KEYS_keys",
        ),
        pytest.param(
            {"CALCULATE_KEYS": {"keys": [{"key": "1"}], "alpha": 3}},
            # The error msg is applied to both CALCULATE_KEYS and UPDATE_KEYS
            ["Missing key: std_cutoff"],
            id="invalid_missing_required_CALCULATE_KEYS_keys",
        ),
    ],
)
def test_invalid_config_setup(test_input, expected_errors):
    schema = build_schema()
    config = configsuite.ConfigSuite(
        test_input,
        schema,
        deduce_required=True,
    )
    assert not config.valid

    msgs = [e.msg for e in config.errors]
    assert len(expected_errors) == len(msgs)
    assert all([any(exp in msg for msg in msgs) for exp in expected_errors])


def test_valid_configuration():
    valid_config_data = {
        "CALCULATE_KEYS": {
            "keys": [{"key": "POLY_OBS"}],
            "alpha": 3,
            "std_cutoff": 0.0001,
        },
        "UPDATE_KEYS": {"keys": [{"key": "POLY_OBS"}]},
    }

    schema = build_schema()
    config = configsuite.ConfigSuite(
        valid_config_data,
        schema,
        deduce_required=True,
    )

    assert config.valid



import os
import shutil
import json
import numpy as np
import pytest
import yaml
from res.enkf import EnKFMain, ResConfig

from semeio.workflows.correlated_observations_scaling import cos
from semeio.workflows.correlated_observations_scaling.cos import (
    CorrelatedObservationsScalingJob,
)
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR
from ert_shared.plugins.plugin_manager import ErtPluginManager


def assert_obs_vector(vector, val_1, index_list=None, val_2=None):
    if index_list is None:
        index_list = []
    for node in vector:
        for index in range(len(node)):
            if index in index_list:
                assert node.getStdScaling(index) == val_2
            else:
                assert node.getStdScaling(index) == val_1


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_ert")
def test_old_enkf_scaling_job(setup_ert):
    res_config = setup_ert
    ert = EnKFMain(res_config)

    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0)

    job = ert.getWorkflowList().getJob("STD_SCALE_CORRELATED_OBS")
    job.run(ert, ["WPR_DIFF_1"])

    assert_obs_vector(obs_vector, np.sqrt(4.0 / 2.0))


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_ert")
def test_installed_python_version_of_enkf_scaling_job(setup_ert, monkeypatch):

    pm = ErtPluginManager(
        plugins=[
            cos,
        ]
    )
    installable_workflow_jobs = pm.get_installable_workflow_jobs()

    res_config = setup_ert
    ert = EnKFMain(res_config)

    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0)

    job_config = {"CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]}}

    with open("job_config.yml", "w") as fout:
        yaml.dump(job_config, fout)

    ert.getWorkflowList().addJob(
        "CORRELATE_OBSERVATIONS_SCALING",
        installable_workflow_jobs["CORRELATED_OBSERVATIONS_SCALING"],
    )

    job = ert.getWorkflowList().getJob("CORRELATE_OBSERVATIONS_SCALING")
    job.run(ert, ["job_config.yml"])

    assert_obs_vector(obs_vector, np.sqrt(4.0 / 2.0))

    job_config["CALCULATE_KEYS"]["keys"][0].update({"index": [400, 800, 1200]})
    with open("job_config.yml", "w") as fout:
        yaml.dump(job_config, fout)
    job.run(ert, ["job_config.yml"])

    assert_obs_vector(
        obs_vector,
        np.sqrt(4.0 / 2.0),
        index_list=[0, 1, 2],
        val_2=np.sqrt(3.0 / 2.0),
    )


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_compare_different_jobs():
    cos_config = {"CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]}}

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)
    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0)

    job = ert.getWorkflowList().getJob("STD_SCALE_CORRELATED_OBS")
    job.run(ert, ["WPR_DIFF_1"])

    # Result of old job:
    assert_obs_vector(obs_vector, np.sqrt(4 / 2))

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    # Result of new job with no sub-indexing:
    assert_obs_vector(obs_vector, np.sqrt(4 / 2))


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_gen_data():
    cos_config = {
        "CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WPR_DIFF_1", "index": [400, 800]}]},
    }

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    obs = ert.getObservations()
    obs_vector = obs["WPR_DIFF_1"]

    assert_obs_vector(obs_vector, 1.0, [0, 1], np.sqrt(4 / 2))

    cos_config["CALCULATE_KEYS"]["keys"][0].update({"index": [400, 800, 1200]})

    CorrelatedObservationsScalingJob(ert).run(cos_config)
    assert_obs_vector(
        obs_vector,
        1.0,
        [0, 1],
        np.sqrt(3.0 / 2.0),
    )

    svd_file = "reports/snake_oil/CorrelatedObservationsScalingJob/svd.json"
    # Assert that data was published correctly
    with open(svd_file) as f:
        svd_reports = json.load(f)
        assert len(svd_reports) == 2

        reported_svd = svd_reports[1]
        assert reported_svd == pytest.approx(
            (
                6.531760256452532,
                2.0045135017540487,
                1.1768827000026516,
            ),
            0.1,
        )

    scale_file = "reports/snake_oil/CorrelatedObservationsScalingJob/scale_factor.json"
    with open(scale_file) as f:
        scalefactor_reports = json.load(f)
        assert len(scalefactor_reports) == 2

        reported_scalefactor = scalefactor_reports[1]
        assert reported_scalefactor == pytest.approx(1.224744871391589, 0.1)


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_summary_data_calc():
    cos_config = {
        "CALCULATE_KEYS": {"keys": [{"key": "WOPR_OP1_108"}, {"key": "WOPR_OP1_144"}]}
    }

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)
    obs = ert.getObservations()

    obs_vector = obs["WOPR_OP1_108"]

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == np.sqrt(1.0)


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.equinor_test
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_summary_data_update():
    cos_config = {
        "CALCULATE_KEYS": {"keys": [{"key": "WWCT:OP_1"}, {"key": "WWCT:OP_2"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WWCT:OP_2", "index": [1, 2, 3, 4, 5]}]},
    }

    test_data_dir = os.path.join(TEST_DATA_DIR, "Equinor", "config", "obs_testing")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("config")
    ert = EnKFMain(res_config)
    obs = ert.getObservations()
    obs_vector = obs["WWCT:OP_2"]

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    for index, node in enumerate(obs_vector):
        if index in cos_config["UPDATE_KEYS"]["keys"][0]["index"]:
            assert node.getStdScaling(index) == np.sqrt(61.0 * 2.0)
        else:
            assert node.getStdScaling(index) == 1.0

    obs_vector = obs["WWCT:OP_1"]

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.equinor_test
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_block_data_calc():
    cos_config = {"CALCULATE_KEYS": {"keys": [{"key": "RFT3"}]}}

    test_data_dir = os.path.join(TEST_DATA_DIR, "Equinor", "config", "with_RFT")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("config")
    ert = EnKFMain(res_config)
    obs = ert.getObservations()

    obs_vector = obs["RFT3"]

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 2.0


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.equinor_test
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_block_and_summary_data_calc():
    cos_config = {"CALCULATE_KEYS": {"keys": [{"key": "FOPT"}, {"key": "RFT3"}]}}

    test_data_dir = os.path.join(TEST_DATA_DIR, "Equinor", "config", "with_RFT")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("config")
    ert = EnKFMain(res_config)
    obs = ert.getObservations()

    obs_vector = obs["RFT3"]

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == 1.0

    CorrelatedObservationsScalingJob(ert).run(cos_config)

    for index, node in enumerate(obs_vector):
        assert node.getStdScaling(index) == np.sqrt(64)



import os
import re
import shutil

import configsuite
import pytest
from res.enkf import EnKFMain, ResConfig
from semeio.workflows.correlated_observations_scaling import job_config
from semeio.workflows.correlated_observations_scaling.exceptions import ValidationError
from semeio.workflows.correlated_observations_scaling.obs_utils import (
    _wildcard_to_dict_list,
    create_active_lists,
    find_and_expand_wildcards,
    keys_with_data,
)
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR


@pytest.mark.parametrize(
    "matching_keys,entry,expected_result",
    [
        (["a_key"], {"key": "a_*"}, [{"key": "a_key"}]),
        (["a_key", "b_key"], {"key": "*key"}, [{"key": "a_key"}, {"key": "b_key"}]),
        (
            ["a_key"],
            {"key": "a_*", "index": [1, 2]},
            [{"key": "a_key", "index": [1, 2]}],
        ),
    ],
)
def test_wildcard_to_dict_list(matching_keys, entry, expected_result):
    assert _wildcard_to_dict_list(matching_keys, entry) == expected_result


def test_find_and_expand_wildcards():
    expected_dict = {
        "ANOTHER_KEY": "something",
        "CALCULATE_KEYS": {
            "keys": [
                {"key": "WOPR_OP1_108"},
                {"key": "WOPR_OP1_144"},
                {"key": "WOPR_OP1_190"},
                {"key": "WOPR_OP1_9"},
                {"key": "WOPR_OP1_36"},
                {"key": "WOPR_OP1_72"},
                {"key": "FOPR"},
            ]
        },
        "UPDATE_KEYS": {
            "keys": [
                {"key": "WOPR_OP1_108"},
                {"key": "WOPR_OP1_144"},
                {"key": "WOPR_OP1_190"},
                {"key": "FOPR"},
            ]
        },
    }

    user_config = {
        "ANOTHER_KEY": "something",
        "CALCULATE_KEYS": {"keys": [{"key": "WOPR_*"}, {"key": "FOPR"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WOPR_OP1_1*"}, {"key": "FOPR"}]},
    }

    observation_list = [
        "WOPR_OP1_108",
        "WOPR_OP1_144",
        "WOPR_OP1_190",
        "WOPR_OP1_9",
        "WOPR_OP1_36",
        "WOPR_OP1_72",
        "FOPR",
    ]

    result_dict = find_and_expand_wildcards(observation_list, user_config)

    assert result_dict == expected_dict


@pytest.mark.parametrize(
    "config_dict,obs_list,expected_fails,err_msg",
    [
        (
            {"CALCULATE_KEYS": {"keys": [{"key": "FOP*"}]}},
            ["FOPR"],
            False,
            None,
        ),
        (
            {"CALCULATE_KEYS": {"keys": [{"key": "TYPO*"}, {"key": "FOP*"}]}},
            ["FOPR"],
            True,
            "Invalid CALCULATE_KEYS TYPO* had no match",
        ),
        (
            {"CALCULATE_KEYS": {"keys": [{"key": "TYPO*"}]}},
            ["FOPR"],
            True,
            "Invalid CALCULATE_KEYS TYPO* had no match",
        ),
        (
            {
                "CALCULATE_KEYS": {"keys": [{"key": "FOPR*"}]},
                "UPDATE_KEYS": {"keys": [{"key": "TYPO*"}]},
            },
            ["FOPR"],
            True,
            "Invalid UPDATE_KEYS TYPO* had no match",
        ),
        (
            {"CALCULATE_KEYS": {"keys": [{"key": "TYPO*"}, {"key": "FOPR"}]}},
            ["FOPR"],
            True,
            "Invalid CALCULATE_KEYS TYPO* had no match",
        ),
    ],
)
def test_failed_wildcard_expansion(config_dict, obs_list, expected_fails, err_msg):
    if expected_fails:
        with pytest.raises(ValidationError) as excinfo:
            find_and_expand_wildcards(obs_list, config_dict)
        # ignore tabs, newlines
        exc_msg = re.sub("\\s+", " ", str(excinfo.value))
        assert exc_msg == err_msg
    else:
        try:
            find_and_expand_wildcards(obs_list, config_dict)
        except ValueError:
            pytest.fail("unexpectedly raised with config: {}".format(config_dict))


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_ert")
def test_create_observation_vectors(setup_ert):

    valid_config_data = {
        "CALCULATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]},
        "UPDATE_KEYS": {"keys": [{"key": "WPR_DIFF_1"}]},
    }
    config = configsuite.ConfigSuite(
        valid_config_data, job_config.build_schema(), deduce_required=True
    )

    res_config = setup_ert
    ert = EnKFMain(res_config)
    obs = ert.getObservations()

    new_events = create_active_lists(obs, config.snapshot.UPDATE_KEYS.keys)

    keys = [event.key for event in new_events]

    assert "WPR_DIFF_1" in keys
    assert "SNAKE_OIL_WPR_DIFF" not in keys


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_add_observation_vectors():

    valid_config_data = {"UPDATE_KEYS": {"keys": [{"key": "WOPR_OP1_108"}]}}

    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(valid_config_data, schema, deduce_required=True)

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil_field")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")

    ert = EnKFMain(res_config)

    obs = ert.getObservations()

    new_events = create_active_lists(obs, config.snapshot.UPDATE_KEYS.keys)

    keys = [event.key for event in new_events]

    assert "WOPR_OP1_108" in keys
    assert "WOPR_OP1_144" not in keys


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_validate_failed_realizations():
    """
    Config has several failed realisations
    """
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "custom_kw")
    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("mini_fail_config")
    ert = EnKFMain(res_config)
    observations = ert.getObservations()

    result = keys_with_data(
        observations,
        ["GEN_PERLIN_1"],
        ert.getEnsembleSize(),
        ert.getEnkfFsManager().getCurrentFileSystem(),
    )
    assert result == ["GEN_PERLIN_1"]


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_validate_no_realizations():
    """
    Ensamble has not run
    """
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "poly_normal")
    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("poly.ert")
    ert = EnKFMain(res_config)
    observations = ert.getObservations()

    result = keys_with_data(
        observations,
        ["POLY_OBS"],
        ert.getEnsembleSize(),
        ert.getEnkfFsManager().getCurrentFileSystem(),
    )
    assert result == []



import pytest
from semeio.jobs.csv_export2 import csv_export2

from tests.jobs.csv_export2 import conftest

test_header = "ENSEMBLE,REAL,DATE,FOPR,X,FLUID_PARAMS:SWL,FLUID_PARAMS:SGCR,A,H\n"


@pytest.mark.skipif(
    conftest.find_available_test_data() is None, reason="no ert-statoil test-data"
)
@pytest.mark.usefixtures("ert_statoil_test_data")
def test_failed_realization_no_summary_file(ert_statoil_test_data):
    path_file = "pathfile_path_to_failed_realizations.txt"
    export_file = "export.txt"
    csv_export2.csv_exporter(
        runpathfile=path_file,
        time_index="monthly",
        outputfile=export_file,
        column_keys="FOPR",
    )
    verifyExportedFile(
        export_file,
        test_header,
        {("iter-1", 1), ("iter-2", 2), ("iter-1", 0), ("iter-1", 2)},
    )


@pytest.mark.skipif(
    conftest.find_available_test_data() is None, reason="no ert-statoil test-data"
)
@pytest.mark.usefixtures("ert_statoil_test_data")
def test_one_iteration(ert_statoil_test_data):
    path_file = "pathfile.txt"
    export_file = "export.txt"
    csv_export2.csv_exporter(
        runpathfile=path_file,
        time_index="monthly",
        outputfile=export_file,
        column_keys="FOPR",
    )
    verifyExportedFile(
        export_file, test_header, {("iter-1", 0), ("iter-1", 1), ("iter-1", 2)}
    )


@pytest.mark.skipif(
    conftest.find_available_test_data() is None, reason="no ert-statoil test-data"
)
@pytest.mark.usefixtures("ert_statoil_test_data")
def test_missing_realization(ert_statoil_test_data):
    path_file = "pathfile2.txt"
    export_file = "export.txt"
    csv_export2.csv_exporter(
        runpathfile=path_file,
        time_index="monthly",
        outputfile=export_file,
        column_keys="FOPR",
    )
    verifyExportedFile(export_file, test_header, {("iter-1", 2), ("iter-1", 0)})


@pytest.mark.skipif(
    conftest.find_available_test_data() is None, reason="no ert-statoil test-data"
)
@pytest.mark.usefixtures("ert_statoil_test_data")
def test_iterations(ert_statoil_test_data):
    path_file = "pathfile3.txt"
    export_file = "export.txt"
    csv_export2.csv_exporter(
        runpathfile=path_file,
        time_index="monthly",
        outputfile=export_file,
        column_keys="FOPR",
    )
    verifyExportedFile(
        export_file,
        test_header,
        {
            ("iter-1", 0),
            ("iter-1", 1),
            ("iter-1", 2),
            ("iter-2", 0),
            ("iter-2", 1),
            ("iter-2", 2),
        },
    )


@pytest.mark.skipif(
    conftest.find_available_test_data() is None, reason="no ert-statoil test-data"
)
@pytest.mark.usefixtures("ert_statoil_test_data")
def test_no_iterations(ert_statoil_test_data):
    path_file = "pathfile1.txt"
    export_file = "export.txt"

    with pytest.raises(KeyError):
        csv_export2.csv_exporter(
            runpathfile=path_file,
            time_index="monthly",
            outputfile=export_file,
            column_keys="FOPR",
        )


def verifyExportedFile(exported_file_name, result_header, result_iter_rel):
    with open(exported_file_name, "r") as exported_file:
        lines = exported_file.readlines()

        assert sorted(lines[0]) == sorted(result_header)

        iter_rel_no = set()

        for index, line in enumerate(lines[1:]):
            tokens = line.split(",")
            iteration, rel = tokens[0], int(tokens[1])

            if (iteration, rel) not in iter_rel_no:
                iter_rel_no.add((iteration, rel))

        assert iter_rel_no == result_iter_rel



class ValidationError(ValueError):
    def __init__(self, message, errors):
        super().__init__(message)
        self.errors = errors
        self.message = message

    def __str__(self):
        msg = "{}\n".format(self.message)
        for error in self.errors:
            path = ".".join(error.key_path) if len(error.key_path) > 0 else "root level"
            msg += "  - {} ({})\n".format(error.msg, path)
        return msg



import yaml
import configsuite

from ert_data.measured import MeasuredData
from ert_shared.libres_facade import LibresFacade
from ert_shared.plugins.plugin_manager import hook_implementation
from semeio.communication import SemeioScript

from semeio.workflows import misfit_preprocessor
from semeio.workflows.correlated_observations_scaling.cos import (
    CorrelatedObservationsScalingJob,
)
from semeio.workflows.correlated_observations_scaling.exceptions import (
    EmptyDatasetException,
)


class MisfitPreprocessorJob(SemeioScript):  # pylint: disable=too-few-public-methods
    def run(self, *args):
        config_record = _fetch_config_record(args)
        measured_record = _load_measured_record(self.ert())
        scaling_configs = misfit_preprocessor.run(
            **{
                "misfit_preprocessor_config": config_record,
                "measured_data": measured_record,
                "reporter": self.reporter,
            }
        )

        # The execution of COS should be moved into
        # misfit_preprocessor.run when COS no longer depend on self.ert
        # to run.
        scaling_params = _fetch_scaling_parameters(config_record, measured_record)
        for scaling_config in scaling_configs:
            scaling_config["CALCULATE_KEYS"].update(scaling_params)

        try:
            CorrelatedObservationsScalingJob(self.ert()).run(scaling_configs)
        except EmptyDatasetException:
            pass


def _fetch_scaling_parameters(config_record, measured_data):
    config = misfit_preprocessor.assemble_config(
        config_record,
        measured_data,
    )
    if not config.valid:
        # The config is loaded by misfit_preprocessor.run first. The
        # second time should never fail!
        raise ValueError("Misfit preprocessor config not valid on second load")

    scale_conf = config.snapshot.scaling
    return {
        "threshold": scale_conf.threshold,
    }


def _fetch_config_record(args):
    if len(args) == 0:
        return {}
    elif len(args) == 1:
        with open(args[0]) as f:
            return yaml.safe_load(f)
    else:
        raise ValueError(
            (
                "Excepted at most one argument, namely the path to a "
                "configuration file. Received {} arguments: {}"
            ).format(len(args), args)
        )


def _load_measured_record(enkf_main):
    facade = LibresFacade(enkf_main)
    obs_keys = [
        facade.get_observation_key(nr) for nr, _ in enumerate(facade.get_observations())
    ]
    return MeasuredData(facade, obs_keys)


@hook_implementation
def legacy_ertscript_workflow(config):
    workflow = config.add_workflow(MisfitPreprocessorJob, "MISFIT_PREPROCESSOR")
    _schema = misfit_preprocessor.config._SCHEMA
    rst_doc = configsuite.docs.generate(_schema)
    workflow.description = rst_doc



import itertools
import json
import os
import pytest
import pandas as pd
from semeio.communication import FileReporter


@pytest.mark.parametrize(
    "msg",
    ["msg", "This is a message!", "And this is another one"],
)
def test_file_reporter_publish_msg_valid_msg(msg, tmpdir):
    tmpdir.chdir()
    reporter = FileReporter(os.getcwd())

    namespace = "my_msg.log"
    reporter.publish_msg(namespace, msg)

    with open(namespace) as f:
        loaded_msg = f.readlines()

    assert (msg + "\n",) == tuple(loaded_msg)


@pytest.mark.parametrize(
    "messages",
    [
        ("msg",),
        ("msg", "This is a message!"),
        ("msg", "This is a message!", "And this is another one"),
    ],
)
def test_file_reporter_publish_msg_multiple_messages(messages, tmpdir):
    tmpdir.chdir()
    reporter = FileReporter(os.getcwd())

    namespace = "my_msg.log"

    for msg in messages:
        reporter.publish_msg(namespace, msg)

    with open(namespace) as f:
        loaded_msg = f.readlines()

    assert tuple(msg + "\n" for msg in messages) == tuple(loaded_msg)


def test_file_reporter_publish_msg_multiple_namespaces(tmpdir):
    tmpdir.chdir()
    namespace1 = "namespace1"
    namespace2 = "namespace2"

    reporter = FileReporter(os.getcwd())

    msg1 = "This is a statement!"
    msg2 = "This is another a statement!"
    reporter.publish_msg(namespace1, msg1)
    reporter.publish_msg(namespace2, msg2)

    with open(namespace1) as f:
        namespace1_msg = f.readlines()
    assert (msg1 + "\n",) == tuple(namespace1_msg)

    with open(namespace2) as f:
        namespace2_msg = f.readlines()
    assert (msg2 + "\n",) == tuple(namespace2_msg)


@pytest.mark.parametrize(
    ("folder", "namespace"),
    list(
        itertools.product(
            ("", "monkey", os.path.join("a", "b")),
            ("namespace", "my_messages.log"),
        )
    ),
)
def test_file_reporter_publisg_msg_output_location(folder, namespace, tmpdir):
    tmpdir.chdir()
    reporter = FileReporter(os.path.join(os.getcwd(), folder))

    msg = "The most arbitrary message there is"
    reporter.publish_msg(namespace, msg)

    expected_output_file = os.path.join(os.getcwd(), folder, namespace)

    with open(expected_output_file) as f:
        loaded_msg = f.readlines()

    assert (msg + "\n",) == tuple(loaded_msg)


@pytest.mark.parametrize(
    "namespace",
    (
        "fdsm/fdsfds",
        "/ds/",
        "/..",
    ),
)
def test_file_reporter_publish_msg_invalid_namespace(namespace, tmpdir):
    tmpdir.chdir()
    reporter = FileReporter(os.getcwd())

    with pytest.raises(ValueError) as err_info:
        reporter.publish_msg(namespace, "")

    assert "Namespace contains path separators" in str(err_info.value)


def test_file_reporter_publish_msg_invalid_output_dir(tmpdir):
    tmpdir.chdir()

    output_dir = os.path.join(os.getcwd(), "output")
    with open(output_dir, "w") as f:
        f.write("You shall not be a directory")

    reporter = FileReporter(output_dir)
    with pytest.raises(ValueError) as err_info:
        reporter.publish_msg("namespace", "")

    assert "Expected output_dir to be a directory" in str(err_info.value)


@pytest.mark.parametrize(
    "data",
    (
        "",
        None,
        [1, 2, "cat"],
        {"a": [1.0, 2], "b": "something"},
    ),
)
def test_file_reporter_publish_valid_json(data, tmpdir):
    tmpdir.chdir()
    namespace = "data"

    reporter = FileReporter(os.getcwd())
    reporter.publish(namespace, data)

    with open(namespace + ".json") as f:
        loaded_data = json.load(f)

    assert loaded_data == [data]


def test_file_reporter_publish_invalid_json(tmpdir):
    tmpdir.chdir()
    namespace = "data"
    data = json  # The json module is not JSON serializable...

    reporter = FileReporter(os.getcwd())

    with pytest.raises(TypeError):
        reporter.publish(namespace, data)


def test_file_reporter_publish_multiple_json(tmpdir):
    tmpdir.chdir()
    namespace = "some_data"
    reporter = FileReporter(os.getcwd())

    data = [0, [0, 12], "Dear JSON"]
    for idx, data_elem in enumerate(data):
        reporter.publish(namespace, data_elem)

        with open(namespace + ".json") as f:
            loaded_data = json.load(f)

        assert loaded_data == data[: idx + 1]


def test_file_reporter_publish_multiple_namespaces(tmpdir):
    tmpdir.chdir()
    namespace1 = "namespace1"
    namespace2 = "namespace2"

    data1 = "This is message number 1"
    data2 = "This is yet another message"

    reporter = FileReporter(os.getcwd())

    reporter.publish(namespace1, data1)
    reporter.publish(namespace2, data2)

    with open(namespace1 + ".json") as f:
        loaded_data1 = json.load(f)
    assert loaded_data1 == [data1]

    with open(namespace2 + ".json") as f:
        loaded_data2 = json.load(f)
    assert loaded_data2 == [data2]


@pytest.mark.parametrize(
    ("folder", "namespace"),
    list(
        itertools.product(
            ("", "monkey", os.path.join("a", "b")),
            ("namespace", "my_messages.log"),
        )
    ),
)
def test_file_reporter_publish_output_location(folder, namespace, tmpdir):
    tmpdir.chdir()
    reporter = FileReporter(os.path.join(os.getcwd(), folder))

    data = [1, 2, 1, 2, "Testing the most arbitrary data there is"]
    reporter.publish(namespace, data)

    expected_output_file = os.path.join(os.getcwd(), folder, namespace) + ".json"

    with open(expected_output_file) as f:
        loaded_data = json.load(f)
    assert [data] == loaded_data


@pytest.mark.parametrize(
    "namespace",
    (
        "fdsm/fdsfds",
        "/ds/",
        "/..",
    ),
)
def test_file_reporter_publish_invalid_namespace(namespace, tmpdir):
    tmpdir.chdir()
    reporter = FileReporter(os.getcwd())

    with pytest.raises(ValueError) as err_info:
        reporter.publish(namespace, "")

    assert "Namespace contains path separators" in str(err_info.value)


def test_file_reporter_publish_invalid_output_dir(tmpdir):
    tmpdir.chdir()

    output_dir = os.path.join(os.getcwd(), "output")
    with open(output_dir, "w") as f:
        f.write("You shall not be a directory")

    reporter = FileReporter(output_dir)
    with pytest.raises(ValueError) as err_info:
        reporter.publish("namespace", "")

    assert "Expected output_dir to be a directory" in str(err_info.value)


@pytest.mark.parametrize(
    "output_dir",
    (
        "fdsm/fdsfds",
        ".",
        "..",
        "reports",
    ),
)
def test_file_reporter_relative_output_dir(output_dir, tmpdir):
    tmpdir.chdir()
    with pytest.raises(ValueError) as err_info:
        FileReporter(output_dir)

    assert "Expected output_dir to be an absolute path" in str(err_info.value)


@pytest.mark.parametrize(
    "data",
    (pd.DataFrame([[1, 2, 3], [2, 3, 5]]),),
)
def test_file_reporter_publish_valid_csv(data, tmpdir):
    tmpdir.chdir()
    namespace = "data"

    reporter = FileReporter(os.getcwd())
    reporter.publish_csv(namespace, data)

    loaded_data = pd.read_csv(namespace + ".csv", index_col=0, header=0)

    assert (loaded_data.values == data.values).all()


def test_file_reporter_publish_invalid_csv(tmpdir):
    tmpdir.chdir()
    namespace = "data"
    data = "not a dataframe"

    reporter = FileReporter(os.getcwd())

    with pytest.raises(AttributeError):
        reporter.publish_csv(namespace, data)



# -*- coding: utf-8 -*-
import itertools
import logging
from scipy.cluster.hierarchy import fcluster, linkage


def spearman_job(
    measured_data,
    threshold,
    reporter,
    criterion="inconsistent",
    depth=2,
    method="single",
    metric="euclidean",
):
    """
    Given measured_data and threshold, it returns configurations that describe
    scaling of these data.
    """
    measured_data.remove_failed_realizations()
    measured_data.remove_inactive_observations()
    measured_data.filter_ensemble_std(1.0e-6)

    simulated_data = measured_data.get_simulated_data()

    correlation_matrix = _calculate_correlation_matrix(simulated_data)
    reporter.publish_csv("correlation_matrix", correlation_matrix)

    clusters = _cluster_analysis(
        correlation_matrix, threshold, criterion, depth, method, metric
    )

    columns = correlation_matrix.columns

    # Here the clusters are joined with the key and data index
    # to group the observations, the column level values are the column
    # headers, where key_index is the observation key and data_index
    # is a range.
    data = list(
        zip(
            clusters,
            columns.get_level_values(0),
            columns.get_level_values("data_index"),
        )
    )

    clustered_data = _remove_singular_obs(_cluster_data(data))
    reporter.publish("clusters", clustered_data)
    job_configs = _config_creation(clustered_data)

    for cluster, val in clustered_data.items():
        logging.info("Cluster nr: {}, clustered data: {}".format(cluster, val))

    return job_configs


def _cluster_data(data):
    groups = {}
    for (nr, key), cluster_group in itertools.groupby(
        sorted(data), key=lambda x: (x[0], x[1])
    ):
        if nr not in groups:
            groups[nr] = {}
        groups[nr].update({key: [index for _, _, index in cluster_group]})
    return groups


def _remove_singular_obs(clusters):
    """Removes clusters with a singular observation."""
    new_cluster_index = 0
    multiobs_clusters = {}
    for _, cluster in clusters.items():
        if sum(map(len, cluster.values())) > 1:
            multiobs_clusters[new_cluster_index] = cluster
            new_cluster_index += 1
        else:
            logging.info(
                "Removed cluster with singular observation: {}".format(cluster)
            )
    return multiobs_clusters


def _config_creation(clusters):
    config = []
    for cluster_nr, cluster in clusters.items():
        config.append(
            {
                "CALCULATE_KEYS": {
                    "keys": [{"key": key, "index": val} for key, val in cluster.items()]
                }
            }
        )
    return config


def _calculate_correlation_matrix(data):
    # Spearman correlation is quite slow, but will be improved in a future version
    # of pandas (https://github.com/pandas-dev/pandas/pull/28151), for now this is
    # equivalent:
    return data.rank().corr(method="pearson")


def _cluster_analysis(correlation_matrix, threshold, criterion, depth, method, metric):
    a = linkage(correlation_matrix, method, metric)
    return fcluster(a, threshold, criterion=criterion, depth=depth)



import configsuite
import pytest
import rstcheck

from semeio.workflows.correlated_observations_scaling import cos, job_config


@pytest.mark.parametrize(
    "input_config",
    [
        cos._calc_keys_example,
        cos._calc_keys_w_index,
        cos._wildcard_example,
        cos._all_keys_w_index,
        *cos._groups_example,
    ],
)
def test_valid_examples(input_config):
    default_values = {
        "CALCULATE_KEYS": {"std_cutoff": 1.0e-6, "alpha": 3},
        "UPDATE_KEYS": {},
    }
    schema = job_config.build_schema()
    config = configsuite.ConfigSuite(
        input_config,
        schema,
        layers=(default_values,),
        deduce_required=True,
    )
    assert config.valid


@pytest.mark.parametrize("input_rst", [cos._DESCRIPTION, cos._EXAMPLES])
def test_valid_rst(input_rst):
    """
    Check that the documentation passed through the plugin system is
    valid rst
    """
    assert not list(rstcheck.check(input_rst))



import os
import datetime
import argparse

from ecl.rft import EclRFTFile
from ecl.grid import EclGrid


def strip_comments(line):
    return line.partition("--")[0].rstrip()


def existing_directory(path):
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError(
            "The path {} is not an existing directory".format(path)
        )
    return path


def load_and_parse_well_time_file(filename):
    """
    The file must be on the format of <STRING INT INT INT INT> which
    represents <well_name day month year report_step>

    Parameters
    ----------
    filename : string
        Filename to open

    Returns
    -------
    [Tuple]
        Returns a list of tuples with (well, datetime, report_step)
    """

    if not os.path.isfile(filename):
        raise argparse.ArgumentTypeError("The path {} does not exist".format(filename))

    well_times = []
    base_error_msg = (
        "Line {line_number} in well_and_time_file {filename} not on proper format: "
    )

    with open(filename) as well_time_file:
        lines = well_time_file.readlines()

    well_time_lines = [(strip_comments(l), i + 1) for i, l in enumerate(lines)]

    for line, line_number in well_time_lines:
        tokens = line.split()

        if not tokens:
            continue

        if len(tokens) != 5:
            err_msg = (
                base_error_msg
                + "Unexpected number of tokens: expected 5 got {num_tokens}"
            )
            raise argparse.ArgumentTypeError(
                err_msg.format(
                    line_number=line_number, filename=filename, num_tokens=len(tokens)
                )
            )

        well = tokens[0]

        try:
            report_step = int(tokens[4])
        except ValueError:
            err_msg = base_error_msg + "Unable to convert {report_step} to int"
            raise argparse.ArgumentTypeError(
                err_msg.format(
                    line_number=line_number, filename=filename, report_step=tokens[4]
                )
            )

        try:
            year = int(tokens[3])
            month = int(tokens[2])
            day = int(tokens[1])
            time = datetime.date(year, month, day)

        except ValueError:
            err_msg = base_error_msg + (
                "Unable to parse date, expected day month year got: "
                "{day} {month} {year}"
            )
            raise argparse.ArgumentTypeError(
                err_msg.format(
                    line_number=line_number,
                    filename=filename,
                    day=tokens[1],
                    month=tokens[2],
                    year=tokens[3],
                )
            )

        well_times.append((well, time, report_step))

    return well_times


def valid_eclbase(file_path):
    """
    The filename is assumed to be without extension and two files
    must be present, <filename>.RFT and <filename>.EGRID.
    Loads both files with respective loaders and returns them

    Parameters
    ----------
    filename : string
        Filename to open

    Returns
    -------
    Tuple
        Returns a tuple with an ecl grid instance and an rft instance
    """
    rft_filepath = file_path + ".RFT"
    if not os.path.isfile(rft_filepath):
        raise argparse.ArgumentTypeError(
            "The path {} does not exist".format(rft_filepath)
        )

    try:
        ecl_rft = EclRFTFile(rft_filepath)
    except (IOError, OSError) as err:
        raise argparse.ArgumentTypeError(
            (
                "Could not load eclipse RFT from file: {fname}\n"
                "With the following error:"
                "\n{ecl_err}"
            ).format(fname=rft_filepath, ecl_err=err)
        )

    grid_filepath = file_path + ".EGRID"
    if not os.path.isfile(grid_filepath):
        raise argparse.ArgumentTypeError(
            "The path {} does not exist".format(grid_filepath)
        )

    try:
        ecl_grid = EclGrid(grid_filepath)
    except (IOError, OSError) as err:
        raise argparse.ArgumentTypeError(
            (
                "Could not load eclipse Grid from file: {fname}\n"
                "With the following error:\n"
                "{ecl_err}"
            ).format(fname=grid_filepath, ecl_err=err)
        )

    return ecl_grid, ecl_rft



import os
import shutil

import pytest
from res.enkf import ResConfig

TEST_DATA_DIR = os.environ.get("LIBRES_TEST_DATA_DIR")


def has_equinor_test_data():
    return os.path.isdir(os.path.join(TEST_DATA_DIR, "Equinor"))


def pytest_runtest_setup(item):
    if item.get_closest_marker("equinor_test") and not has_equinor_test_data():
        pytest.skip("Test requires Equinor data")


@pytest.fixture()
def setup_ert(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")
    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    yield res_config

    os.chdir(cwd)


@pytest.fixture()
def setup_tmpdir(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    yield
    os.chdir(cwd)



import pandas as pd
from fmu import ensemble


def csv_exporter(runpathfile, time_index, outputfile, column_keys=None):
    """Export CSV data (summary and parameters) from an EnsembleSet

    The EnsembleSet is described by a runpathfile which must exists
    and point to realizations"""
    ensemble_set = ensemble.EnsembleSet(
        name="ERT EnsembleSet for CSV_EXPORT2", runpathfile=runpathfile
    )
    summary = ensemble_set.load_smry(time_index=time_index, column_keys=column_keys)
    parameters = ensemble_set.parameters
    summary_parameters = pd.merge(summary, parameters)
    summary_parameters.to_csv(outputfile, index=False)



import json
import os


class FileReporter(object):
    def __init__(self, output_dir):
        self._output_dir = output_dir

        if not os.path.isabs(output_dir):
            err_fmt = "Expected output_dir to be an absolute path, received '{}'"
            raise ValueError(err_fmt.format(output_dir))

    def publish_csv(self, namespace, data):
        output_file = self._prepare_output_file(namespace) + ".csv"
        data.to_csv(output_file)

    def publish(self, namespace, data):
        output_file = self._prepare_output_file(namespace) + ".json"

        if os.path.exists(output_file):
            with open(output_file) as f:
                all_data = json.load(f)
        else:
            all_data = []

        all_data.append(data)
        with open(output_file, "w") as f:
            json.dump(all_data, f)

    def publish_msg(self, namespace, msg):
        fmt = "{}\n"
        output_file = self._prepare_output_file(namespace)
        with open(output_file, "a") as f:
            f.write(fmt.format(msg))

    def _prepare_output_file(self, namespace):
        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir)

        if not os.path.isdir(self._output_dir):
            msg = "Expected output_dir to be a directory ({})"
            raise ValueError(msg.format(self._output_dir))

        if os.path.sep in namespace:
            err_msg = "Namespace contains path separators ({})"
            raise ValueError(err_msg.format(namespace))

        return os.path.join(self._output_dir, namespace)



import argparse

from ert_data.measured import MeasuredData
from ert_shared.libres_facade import LibresFacade
from ert_shared.plugins.plugin_manager import hook_implementation
from semeio.communication import SemeioScript
from semeio.workflows.correlated_observations_scaling.exceptions import (
    EmptyDatasetException,
)
from semeio.workflows.correlated_observations_scaling.cos import (
    CorrelatedObservationsScalingJob,
)
from semeio.workflows.spearman_correlation_job.job import spearman_job


class SpearmanCorrelationJob(SemeioScript):  # pylint: disable=too-few-public-methods
    def run(self, *args):
        facade = LibresFacade(self.ert())

        obs_keys = [
            facade.get_observation_key(nr)
            for nr, _ in enumerate(facade.get_observations())
        ]

        measured_data = MeasuredData(facade, obs_keys)

        parser = spearman_job_parser()
        args = parser.parse_args(args)

        scaling_configs = spearman_job(measured_data, args.threshold, self.reporter)

        if not args.dry_run:
            try:
                CorrelatedObservationsScalingJob(self.ert()).run(scaling_configs)
            except EmptyDatasetException:
                pass


def spearman_job_parser():
    description = """
    A module that calculates the Spearman correlation in simulated data
    and clusters
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-t",
        "--threshold",
        required=False,
        default=1.15,
        type=float,
        help="""
        Forms flat clusters so that the original
        observations in each flat cluster have no greater a
        cophenetic distance than `t`.
        """,
    )
    parser.add_argument(
        "--output-file",
        required=False,
        type=str,
        help="Name of the outputfile. The format will be yaml.",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        required=False,
        help="Dry run, no scaling will be performed",
        action="store_true",
    )
    return parser


@hook_implementation
def legacy_ertscript_workflow(config):
    workflow = config.add_workflow(SpearmanCorrelationJob, "SPEARMAN_CORRELATION")
    workflow.parser = spearman_job_parser



import os
import shutil

import pytest

TEST_DATA_DIR = os.environ.get("ERT_TESTDATA_ROOT")
TEST_DATA_DIR_BERGEN = "/d/proj/bg/enkf/ErtTestData"
TEST_DATA_DIR_STAVANGER = "/project/res-testdata/ErtTestData"


def find_available_test_data():
    if TEST_DATA_DIR is None:
        if os.path.isdir(TEST_DATA_DIR_STAVANGER):
            return TEST_DATA_DIR_STAVANGER
        elif os.path.isdir(TEST_DATA_DIR_BERGEN):
            return TEST_DATA_DIR_BERGEN

    return TEST_DATA_DIR


def has_equinor_test_data():
    return os.path.isdir(find_available_test_data())


def pytest_runtest_setup(item):
    if item.get_closest_marker("equinor_test") and not has_equinor_test_data():
        pytest.skip("Test requires Equinor data")


@pytest.fixture()
def ert_statoil_test_data(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    test_data_dir = os.path.join(find_available_test_data(), "ert-statoil", "spotfire")
    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    yield

    os.chdir(cwd)


@pytest.fixture()
def setup_tmpdir(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    yield
    os.chdir(cwd)



import logging
import os
import sys
from distutils.dir_util import copy_tree

import pytest

if sys.version_info.major >= 3:
    from semeio.jobs.scripts import design_kw


@pytest.fixture
def input_data(tmpdir):
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
    copy_tree(data_dir, tmpdir.strpath)

    cwd = os.getcwd()
    tmpdir.chdir()

    yield

    os.chdir(cwd)


_templatefile = "template.yml.tmpl"
_resultfile = "result.txt"
_log_level = "DEBUG"
_default_log_level = "WARNING"


def test_argparse(input_data):
    args = [_templatefile, _resultfile]
    parser = design_kw.create_parser()
    res = parser.parse_args(args)

    assert res
    assert res.templatefile == _templatefile
    assert res.resultfile == _resultfile
    assert res.log_level == logging.getLevelName(_default_log_level)


def test_argparse_with_logging(input_data):
    args = [_templatefile, _resultfile, "--log-level", _log_level]
    parser = design_kw.create_parser()
    res = parser.parse_args(args)
    assert res
    assert res.templatefile == _templatefile
    assert res.resultfile == _resultfile
    assert res.log_level == logging.getLevelName(_log_level)


def test_argparse_file_not_exists():
    with pytest.raises(SystemExit):
        design_kw.main(["file_not_existing.yml.tmpl", _resultfile])


def test_argparse_result_file_missing(input_data):
    with pytest.raises(SystemExit):
        design_kw.main([_templatefile])



import os
from distutils.dir_util import copy_tree
import numpy
import pandas as pd

import pytest

from semeio.jobs.scripts.gendata_rft import main_entry_point, _build_parser

ECL_BASE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "data/norne/NORNE_ATW2013"
)
EXPECTED_RESULTS_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "data/norne/expected_result"
)

MOCK_DATA_CONTENT = {
    "RFT_B-1AH_0_active": [1] * 5,
    "RFT_B-4AH_0_active": [1] * 7 + [0] * 1 + [1] * 21,
    "RFT_B-4H_0_active": [1] * 20,
    "RFT_C-1H_0_active": [1] * 21,
    "RFT_C-3H_0_active": [1] * 21,
    "RFT_C-4AH_0_active": [1] * 15,
}


def _assert_almost_equal_line_by_line(file1, file2):
    with open(file1) as fh:
        file1_content = fh.readlines()

    with open(file2) as fh:
        file2_content = fh.readlines()

    assert len(file1_content) == len(file2_content)

    for line1, line2 in zip(file1_content, file2_content):
        try:
            line1, line2 = float(line1), float(line2)
        except ValueError:
            continue
        numpy.testing.assert_almost_equal(line1, line2, decimal=7)


def _generate_mock_data(write_directory):
    for fname, content in MOCK_DATA_CONTENT.items():
        with open(os.path.join(write_directory, fname), "w+") as fh:
            fh.write("\n".join([str(c) for c in content]))


@pytest.fixture
def input_data(tmpdir):
    data_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data/norne/gendata_rft_input_files"
    )
    copy_tree(data_dir, tmpdir.strpath)

    cwd = os.getcwd()
    tmpdir.chdir()

    expected_results_dir = os.path.join(tmpdir.strpath, "expected_results")
    os.mkdir(expected_results_dir)

    copy_tree(EXPECTED_RESULTS_PATH, expected_results_dir)
    _generate_mock_data(expected_results_dir)

    try:
        yield

    finally:
        os.chdir(cwd)


def test_gendata_rft_csv(tmpdir, input_data):
    csv_filename = "gendata_rft.csv"
    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
        "-c",
        csv_filename,
    ]

    main_entry_point(arguments)
    assert os.path.exists(csv_filename)
    dframe = pd.read_csv(csv_filename)
    assert not dframe.empty

    required_columns = {
        "utm_x",
        "utm_y",
        "measured_depth",
        "true_vertical_depth",
        "zone",
        "i",
        "j",
        "k",
        "pressure",
        "valid_zone",
        "is_active",
        "inactive_info",
        "well",
        "time",
    }
    assert required_columns.issubset(set(dframe.columns))
    assert len(dframe["well"].unique()) == len(MOCK_DATA_CONTENT)

    # i-j-k always non-null together:
    assert (dframe["i"].isnull() == dframe["j"].isnull()).all()
    assert (dframe["i"].isnull() == dframe["k"].isnull()).all()

    # inactive_info must be non-null when active = False
    assert (~dframe[~dframe["is_active"]]["inactive_info"].isnull()).all()


def test_gendata_rft_directory(tmpdir, input_data):
    outputdirectory = "rft_output_dir"
    tmpdir.mkdir(outputdirectory)
    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
        "-c",
        "csvfile.csv",
        "-o",
        outputdirectory,
    ]
    main_entry_point(arguments)
    well_count = 6
    files_pr_well = 3
    assert len(list(os.walk(outputdirectory))[0][2]) == well_count * files_pr_well
    assert os.path.exists("csvfile.csv")  # Should be independent of --outputdirectory


def test_gendata_rft_entry_point(tmpdir, input_data):

    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
    ]

    main_entry_point(arguments)
    expected_results_dir = os.path.join(tmpdir.strpath, "expected_results")

    expected_files = [
        os.path.join(EXPECTED_RESULTS_PATH, fname)
        for fname in os.listdir(EXPECTED_RESULTS_PATH)
    ]

    expected_files.extend(
        [
            os.path.join(expected_results_dir, fname)
            for fname in os.listdir(expected_results_dir)
        ]
    )

    for expected_file in expected_files:
        filename = os.path.basename(expected_file)
        assert filename in os.listdir(tmpdir.strpath)

        if not filename.endswith("inactive_info"):
            result_file = os.path.join(tmpdir.strpath, filename)

            _assert_almost_equal_line_by_line(expected_file, result_file)


def test_gendata_inactive_info_point_not_in_grid(tmpdir, input_data):

    with open("B-1AH.txt", "a+") as fh:
        fh.write("0 1 2 3\n")

    with open("well_and_time.txt", "w+") as fh:
        fh.write("B-1AH 1 12 2005 0\n")

    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
    ]
    main_entry_point(arguments)

    with open("RFT_B-1AH_0_inactive_info") as fh:
        result = fh.read()
        assert result.startswith(
            "TRAJECTORY_POINT_NOT_IN_GRID (utm_x=0.0, utm_y=1.0, measured_depth=2.0)"
        )


def test_gendata_inactive_info_zone_mismatch(tmpdir, input_data):

    with open("well_and_time.txt", "w+") as fh:
        fh.write("B-1AH 1 12 2005 0\n")

    with open(os.path.join(tmpdir.strpath, "B-1AH.txt"), "r+") as fh:
        lines = fh.readlines()

    line = lines[-1].rsplit(" ", 1)[0]
    line = line + " last_zone"

    with open("B-1AH.txt", "w+") as fh:
        fh.write(line)

    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
    ]
    main_entry_point(arguments)

    with open("RFT_B-1AH_0_inactive_info") as fh:
        result = fh.read()
        assert result.startswith("ZONE_MISMATCH (utm_x=")


def test_gendata_inactive_info_not_in_rft(tmpdir, input_data):

    with open("well_and_time.txt", "w+") as fh:
        fh.write("B-1AH 1 12 2005 0\n")

    with open(os.path.join(tmpdir.strpath, "B-1AH.txt"), "r+") as fh:
        lines = fh.readlines()

    line = lines[-1].rsplit(" ", 3)[0]
    line += " 2700 2700"

    with open("B-1AH.txt", "w+") as fh:
        fh.write(line)

    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
    ]
    main_entry_point(arguments)

    with open("RFT_B-1AH_0_inactive_info") as fh:
        result = fh.read()
        assert result.startswith("TRAJECTORY_POINT_NOT_IN_RFT (utm_x=")


def test_gendata_inactive_info_zone_missing_value(tmpdir, input_data):

    with open("well_and_time.txt", "w+") as fh:
        fh.write("B-1AH 1 12 2005 0\n")

    with open("zonemap.txt", "w+") as fh:
        fh.write("1 zone1\n")

    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
    ]

    main_entry_point(arguments)

    with open("RFT_B-1AH_0_inactive_info") as fh:
        result = fh.read()
        assert result.startswith("ZONEMAP_MISSING_VALUE (utm_x=")


def test_gendata_partial_rft_file(tmpdir, input_data):
    """Test how the code behaves when some report steps are missing, e.g.
    a Eclipse simulation that has crashed midway.

    We emulate this situation by asking for well-times that are not in the
    binary RFT output in the test dataset.

    In such a situation, we do not want the OK file to written,
    but the CSV file can contain results of data up until the Eclipse crash.
    """

    with open("well_and_time.txt", "a") as file_h:
        # Binary example file does not go as far as 2045:
        file_h.write("B-1AH 1 12 2045 0")

    csv_filename = "gendata_rft.csv"
    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "well_and_time.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
        "-c",
        csv_filename,
    ]

    main_entry_point(arguments)

    assert os.path.exists(csv_filename)
    assert not pd.read_csv(csv_filename).empty
    assert not os.path.exists("GENDATA_RFT.OK")


def test_gendata_rft_empty_well_and_time(tmpdir, input_data):
    def file_count_cwd():
        return len(list(os.walk("."))[0][2])

    with open("empty.txt", "w") as file_h:
        file_h.write("")

    arguments = [
        "-e",
        ECL_BASE,
        "-w",
        "empty.txt",
        "-t",
        tmpdir.strpath,
        "-z",
        "zonemap.txt",
        "-c",
        "notwritten.csv",
    ]

    pre_file_count = file_count_cwd()
    main_entry_point(arguments)

    # Empty file is seen as an error, we should not write the OK file, and
    # there should be no CSV file.

    assert not os.path.exists("notwritten.csv")
    assert not os.path.exists("GENDATA_RFT.OK")
    assert file_count_cwd() == pre_file_count


def test_defaults():
    """Default filename for CSV export and outputdir is/can be handled at multiple
    layers.  ERT users infer the default from the JOB_DESCRIPTION file, while command
    line users (or Everest/ERT3?) might get the default from the argparse setup.

    To avoid confusion, these should be in sync, enforced by this test"""

    # Navigate to the JOB_DESCRIPTION in the source code tree:
    import semeio.jobs

    job_description_file = os.path.join(
        os.path.dirname(semeio.jobs.__file__), "config_jobs", "GENDATA_RFT"
    )

    # Crude parsing of the file
    for line in open(job_description_file).readlines():
        if line.startswith("DEFAULT"):
            if line.split()[0:2] == ["DEFAULT", "<CSVFILE>"]:
                csv_job_default = line.split()[2]
            if line.split()[0:2] == ["DEFAULT", "<OUTPUTDIRECTORY>"]:
                directory_job_default = line.split()[2]

    # And compare with argparse:
    assert csv_job_default == _build_parser().get_default("csvfile")
    assert directory_job_default == _build_parser().get_default("outputdirectory")



#!/usr/bin/env python
#
import argparse
import logging
import sys

from semeio import valid_file
from semeio.jobs.design2params import design2params

description = """
Reads a design matrix in XLSX-format and

* Fetches values belonging to given ERT realization and *appends*
  key-value pairs to parameters.txt
* Converts specified worksheet to txt-format and puts designmatrix.txt in RUNPATH
* Creates designparameters.txt which only contains variables taken from the design
  matrix, this can be given to CUSTOM_KW

Requires a matrix with column header as strings in topmost row in xls-file.
Row 2 in xls-file must then correspond with the data you want for realization-0
Column 1 in xls-file should contain the realization number, and will be ignored
by this script.  You must run this script as a FORWARD_MODEL from your ERT config,
before you run DESIGN_KW.
"""

category = "utility.templating"


def create_parser():
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "realization",
        type=int,
        help="Which realization",
    )

    parser.add_argument(
        "xlsfilename",
        type=valid_file,
        help="Path to design-matrix (xlsx file)",
    )

    parser.add_argument(
        "designsheetname",
        type=str,
        help="Design sheet name",
    )

    parser.add_argument(
        "defaultssheetname",
        type=str,
        nargs="?",
        help="Defaults sheet name",
        default=None,
    )

    parser.add_argument(
        "--parametersfilename",
        "-p",
        required=False,
        default=design2params._PARAMETERS_TXT,
        type=str,
    )

    parser.add_argument(
        "--log-level",
        "-l",
        required=False,
        default="WARNING",
        type=logging.getLevelName,
    )

    return parser


def main(args):
    parser = create_parser()
    parsed_args = parser.parse_args(args)
    design2params.run(
        realization=parsed_args.realization,
        xlsfilename=parsed_args.xlsfilename,
        designsheetname=parsed_args.designsheetname,
        defaultssheetname=parsed_args.defaultssheetname,
        parametersfilename=parsed_args.parametersfilename,
        log_level=parsed_args.log_level,
    )


if __name__ == "__main__":
    main(sys.argv[1:])



import os
import shutil
import json
import pytest
import pandas as pd
import semeio.workflows.spearman_correlation_job.spearman_correlation as sc
from scipy import stats
from res.enkf import EnKFMain, ResConfig
from semeio.workflows.correlated_observations_scaling.exceptions import (
    EmptyDatasetException,
)
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR

from unittest.mock import Mock


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_gen_data(monkeypatch):
    run_mock = Mock()
    scal_job = Mock(return_value=Mock(run=run_mock))
    monkeypatch.setattr(sc, "CorrelatedObservationsScalingJob", scal_job)

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)
    sc.SpearmanCorrelationJob(ert).run(*["-t", "1.0"])

    # call_args represents the clusters, we expect the snake_oil
    # observations to generate this amount of them
    # call_args is a call object, which itself is a tuple of args and kwargs.
    # In this case, we want args, and the first element of the arguments, which
    # again is a tuple containing the configuration which is a list of configs.
    assert len(list(run_mock.call_args)[0][0]) == 47, "wrong number of clusters"

    cor_matrix_file = "reports/snake_oil/SpearmanCorrelationJob/correlation_matrix.csv"

    pd.read_csv(cor_matrix_file, index_col=[0, 1], header=[0, 1])

    clusters_file = "reports/snake_oil/SpearmanCorrelationJob/clusters.json"
    with open(clusters_file) as f:
        cluster_reports = json.load(f)
        assert len(cluster_reports) == 1

        clusters = cluster_reports[0]
        assert len(clusters.keys()) == 47


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_scaling():
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    sc.SpearmanCorrelationJob(ert).run(*["-t", "1.0"])

    # assert that this arbitrarily chosen cluster gets scaled as expected
    obs = ert.getObservations()["FOPR"]
    for index in [13, 14, 15, 16, 17, 18, 19, 20]:
        assert obs.getNode(index).getStdScaling() == 2.8284271247461903


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_skip_clusters_yielding_empty_data_matrixes(monkeypatch):
    def raising_scaling_job(data):
        if data == {"CALCULATE_KEYS": {"keys": [{"index": [88, 89], "key": "FOPR"}]}}:
            raise EmptyDatasetException("foo")

    scaling_mock = Mock(return_value=Mock(**{"run.side_effect": raising_scaling_job}))
    monkeypatch.setattr(sc, "CorrelatedObservationsScalingJob", scaling_mock)

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)
    job = sc.SpearmanCorrelationJob(ert)

    try:
        job.run(*["-t", "1.0"])
    except EmptyDatasetException:
        pytest.fail("EmptyDatasetException was not handled by SC job")


@pytest.fixture()
def facade():
    facade = Mock()
    facade.get_observations.return_value = []
    return facade


@pytest.fixture()
def measured_data():
    r1 = [1, 5, 3]
    r2 = [2, 90, 2]
    r3 = [3, 2, 1]
    measured_data = Mock()
    columns = [("obs_name", n) for n in range(3)]
    data = pd.DataFrame(
        [r1, r2, r3],
        columns=pd.MultiIndex.from_tuples(columns, names=["key_index", "data_index"]),
    )
    measured_data.get_simulated_data.return_value = data
    return measured_data


@pytest.mark.usefixtures("setup_tmpdir")
def test_main_entry_point_syn_data(monkeypatch, facade, measured_data):
    run_mock = Mock()
    scal_job = Mock(return_value=Mock(run=run_mock))
    monkeypatch.setattr(sc, "CorrelatedObservationsScalingJob", scal_job)

    facade_mock = Mock()
    facade_mock.return_value = facade
    md_mock = Mock()
    md_mock.return_value = measured_data

    monkeypatch.setattr(sc, "LibresFacade", facade_mock)
    monkeypatch.setattr(sc, "MeasuredData", md_mock)

    resconfig_mock = Mock()
    resconfig_mock.config_path = "config_path"
    resconfig_mock.user_config_file = "user_config_file"

    ert_mock = Mock(resConfig=Mock(return_value=resconfig_mock))

    sc.SpearmanCorrelationJob(ert_mock).run(*["-t", "1.0"])

    cor_matrix_file = (
        "config_path/reports/user_config_file/"
        "SpearmanCorrelationJob/correlation_matrix.csv"
    )

    r1 = [1, 5, 3]
    r2 = [2, 90, 2]
    r3 = [3, 2, 1]
    expected_corr_matrix = []
    for a in zip(r1, r2, r3):
        row = []
        for b in zip(r1, r2, r3):
            expected_corr = stats.spearmanr(a, b)
            row.append(expected_corr[0])
        expected_corr_matrix.append(row)

    corr_matrix = pd.read_csv(cor_matrix_file, index_col=[0, 1], header=[0, 1])

    assert (expected_corr_matrix == corr_matrix.values).all()
    clusters_file = (
        "config_path/reports/user_config_file/SpearmanCorrelationJob/clusters.json"
    )
    with open(clusters_file) as f:
        cluster_reports = json.load(f)
        assert len(cluster_reports) == 1

        clusters = cluster_reports[0]
        assert len(clusters.keys()) == 1



from semeio.workflows.misfit_preprocessor.exceptions import ValidationError
from semeio.workflows.misfit_preprocessor.config import (
    SPEARMAN_CORRELATION,
    assemble_config,
)
from semeio.workflows.spearman_correlation_job.job import spearman_job


def run(misfit_preprocessor_config, measured_data, reporter):
    config = assemble_config(misfit_preprocessor_config, measured_data)
    if not config.valid:
        raise ValidationError(
            "Invalid configuration of misfit preprocessor", config.errors
        )

    config = config.snapshot
    if config.clustering.method == SPEARMAN_CORRELATION:
        sconfig = config.clustering.spearman_correlation
        scaling_configs = spearman_job(
            measured_data,
            sconfig.fcluster.t,
            reporter,
            criterion=sconfig.fcluster.criterion,
            depth=sconfig.fcluster.depth,
            method=sconfig.linkage.method,
            metric=sconfig.linkage.metric,
        )
    else:
        raise AssertionError(
            "Unknown clustering method: {}".format(config.clustering.method)
        )

    return scaling_configs



import datetime
import logging
import os
import threading
from logging.handlers import BufferingHandler
from pathlib import Path
from types import MethodType

from res.enkf import ErtScript

from semeio.communication.reporter import FileReporter

SEMEIOSCRIPT_LOG_FILE = "workflow-log.txt"


class _LogHandlerContext(object):
    def __init__(self, log, handler):
        self._log = log
        self._handler = handler

    def __enter__(self):
        self._log.addHandler(self._handler)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._log.removeHandler(self._handler)


class _ReportHandler(BufferingHandler):
    def __init__(self, output_dir, thread_id):
        super(_ReportHandler, self).__init__(1)
        self._reporter = FileReporter(output_dir)
        self._namespace = SEMEIOSCRIPT_LOG_FILE
        self.addFilter(lambda rec: rec.thread == thread_id)

    def flush(self):
        for log_record in self.buffer:
            self._reporter.publish_msg(self._namespace, self._format_record(log_record))

        super(_ReportHandler, self).flush()

    def _format_record(self, log_record):
        log_fmt = "{log_level} [{log_time}]: {log_message}"
        return log_fmt.format(
            log_level=log_record.levelname,
            log_time=datetime.datetime.fromtimestamp(log_record.created),
            log_message=log_record.message,
        )


class SemeioScript(ErtScript):  # pylint: disable=too-few-public-methods
    """
    SemeioScript is a workflow utility extending the functionality of ErtScript.
    In particular it provides a `self.reporter` instance available for passing
    data to the common storage. In addition, while `self.run` is being executed
    it forwards log statements to the reporter as well.
    """

    def __init__(self, ert):
        super(SemeioScript, self).__init__(ert)
        self._output_dir = self._get_output_dir()
        self._reporter = FileReporter(self._output_dir)
        self._wrap_run()

    def _wrap_run(self):
        # pylint: disable=access-member-before-definition
        self._real_run = self.run

        def run_with_handler(self, *args, **kwargs):
            log = logging.getLogger("")
            thread_id = threading.get_ident()
            report_handler = _ReportHandler(self._output_dir, thread_id)
            with _LogHandlerContext(log, report_handler):
                self._real_run(*args, **kwargs)

        self.run = MethodType(run_with_handler, self)

    def _get_output_dir(self):
        res_config = self.ert().resConfig()
        base_dir = res_config.config_path
        base_dir = os.path.realpath(base_dir)
        sub_dir = Path(res_config.user_config_file).stem
        return os.path.join(
            base_dir,
            "reports",
            sub_dir,
            type(self).__name__,
        )

    @property
    def reporter(self):
        return self._reporter



import pytest

from semeio.workflows import misfit_preprocessor


@pytest.mark.parametrize(
    "threshold, expected_valid",
    [
        (-1, False),
        (0, False),
        (10, False),
        (1.0001, False),
        (0.001, True),
        (0.5, True),
        (0.999, True),
    ],
)
def test_valid_scaling_threshold(threshold, expected_valid):
    config_data = {
        "scaling": {"threshold": threshold, "std_cutoff": 2, "alpha": 3},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        assert ("scaling", "threshold") == config.errors[0].key_path


def test_default_scaling_threshold():
    config_data = {
        "scaling": {"std_cutoff": 2, "alpha": 3},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    assert config.valid, config.errors
    assert 0.95 == config.snapshot.scaling.threshold


@pytest.mark.parametrize(
    "std_cutoff, expected_valid", [(1, True), (2, True), (None, True)]
)
def test_scaling_std_cutoff(std_cutoff, expected_valid):
    config_data = {
        "scaling": {"threshold": 0.1, "std_cutoff": std_cutoff, "alpha": 3},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        assert ("scaling", "std_cutoff") == config.errors[0].key_path


@pytest.mark.parametrize("alpha, expected_valid", [(1, True), (2, True)])
def test_scaling_alpha(alpha, expected_valid):
    config_data = {
        "scaling": {"alpha": alpha},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        assert ("scaling", "alpha") == config.errors[0].key_path


@pytest.mark.parametrize(
    "clustering_method, expected_valid",
    [("spearman_correlation", True), ("super_clustering", False), (None, False)],
)
def test_clustering_method(clustering_method, expected_valid):
    config_data = {"clustering": {"method": clustering_method}}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        assert ("clustering", "method") == config.errors[0].key_path


def test_default_clustering_method():
    config_data = {}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    assert config.valid, config.errors
    assert "spearman_correlation" == config.snapshot.clustering.method


@pytest.mark.parametrize(
    "threshold, criterion, expected_valid, postfix_key_path",
    [
        (None, "inconsistent", False, ("t",)),
        (None, "distance", False, ("t",)),
        (None, "maxclust", False, ("t",)),
        (None, "monocrit", False, ("t",)),
        (None, "maxclust_monocrit", False, ("t",)),
        (0, "inconsistent", False, ("t",)),
        (0, "distance", False, ("t",)),
        (0, "maxclust", False, ("t",)),
        (0, "monocrit", False, ("t",)),
        (0, "maxclust_monocrit", False, ("t",)),
        (12, "inconsistent", True, ()),
        (12, "distance", True, ()),
        (12, "maxclust", True, ()),
        (12, "monocrit", True, ()),
        (12, "maxclust_monocrit", True, ()),
        (1.2, "inconsistent", True, ()),
        (1.2, "distance", True, ()),
        (1.2, "maxclust", False, ()),
        (1.2, "monocrit", True, ()),
        (1.2, "maxclust_monocrit", False, ()),
    ],
)
def test_valid_spearman_threshold(
    threshold, criterion, expected_valid, postfix_key_path
):
    config_data = {
        "clustering": {
            "spearman_correlation": {
                "fcluster": {"t": threshold, "criterion": criterion},
            },
        },
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        expected_key_path = ("clustering", "spearman_correlation", "fcluster")
        expected_key_path += postfix_key_path
        assert expected_key_path == config.errors[0].key_path


@pytest.mark.parametrize(
    "criterion, default_threshold",
    [
        ("inconsistent", 1.15),
        ("distance", 1.15),
        ("maxclust", 5),
        ("monocrit", 1.15),
        ("maxclust_monocrit", 5),
    ],
)
def test_default_spearman_threshold(criterion, default_threshold):
    config_data = {
        "clustering": {"spearman_correlation": {"fcluster": {"criterion": criterion}}}
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    assert config.valid, config.errors
    assert (
        default_threshold == config.snapshot.clustering.spearman_correlation.fcluster.t
    )


@pytest.mark.parametrize(
    "depth, expected_valid",
    [
        (-10, False),
        (0, False),
        (1, True),
        (10 ** 10, True),
        (None, False),
        (1.5, False),
    ],
)
def test_valid_spearman_depth(depth, expected_valid):
    config_data = {
        "clustering": {"spearman_correlation": {"fcluster": {"depth": depth}}},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        expected_key_path = ("clustering", "spearman_correlation", "fcluster", "depth")
        assert expected_key_path == config.errors[0].key_path


def test_default_spearman_depth():
    config_data = {}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    assert config.valid, config.errors
    assert 2 == config.snapshot.clustering.spearman_correlation.fcluster.depth


@pytest.mark.parametrize(
    "method, expected_valid",
    [
        ("single", True),
        ("complete", True),
        ("average", True),
        ("weighted", True),
        ("centroid", True),
        ("ward", True),
        (None, False),
        ("secret", False),
    ],
)
def test_valid_linkage_method(method, expected_valid):
    config_data = {
        "clustering": {"spearman_correlation": {"linkage": {"method": method}}},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        expected_key_path = ("clustering", "spearman_correlation", "linkage", "method")
        assert expected_key_path == config.errors[0].key_path


def test_default_linkage_method():
    config_data = {}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    assert config.valid, config.errors
    assert "single" == config.snapshot.clustering.spearman_correlation.linkage.method


@pytest.mark.parametrize(
    "metric, expected_valid",
    [
        ("braycurtis", True),
        ("canberra", True),
        ("chebyshev", True),
        ("cityblock", True),
        ("correlation", True),
        ("cosine", True),
        ("dice", True),
        ("euclidean", True),
        ("hamming", True),
        ("jaccard", True),
        ("jensenshannon", True),
        ("kulsinski", True),
        ("mahalanobis", True),
        ("matching", True),
        ("minkowski", True),
        ("rogerstanimoto", True),
        ("russellrao", True),
        ("seuclidean", True),
        ("sokalmichener", True),
        ("sokalsneath", True),
        ("sqeuclidean", True),
        ("yule", True),
        (None, False),
        ("secret", False),
    ],
)
def test_valid_linkage_metric(metric, expected_valid):
    config_data = {
        "clustering": {"spearman_correlation": {"linkage": {"metric": metric}}},
    }
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    if expected_valid:
        assert config.valid, config.errors
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        expected_key_path = ("clustering", "spearman_correlation", "linkage", "metric")
        assert expected_key_path == config.errors[0].key_path


def test_default_linkage_metric():
    config_data = {}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, ("some_observation",)
    )

    assert config.valid, config.errors
    assert "euclidean" == config.snapshot.clustering.spearman_correlation.linkage.metric


@pytest.mark.parametrize(
    "observation_filter, observation_keys, expected_obs, expected_valid",
    [
        (("my_obs",), ("my_obs", "other_obs"), ("my_obs",), True),
        (("my_obs", "obs1"), ("my_obs", "other_obs", "obs1"), ("my_obs", "obs1"), True),
        (("secret_obs",), ("my_obs", "other_obs"), (), False),
        (("*",), ("my_obs", "other_obs"), ("my_obs", "other_obs"), True),
        (
            ("*1", "*2"),
            ("a1", "a2", "a3", "aa1", "aa2", "aa3"),
            ("a1", "a2", "aa1", "aa2"),
            True,
        ),
        (
            ("*1", "*2", "*5"),
            ("a1", "a2", "a3", "aa1", "aa2", "aa3"),
            ("a1", "a2", "aa1", "aa2"),
            False,
        ),
    ],
)
def test_valid_observations(
    observation_filter,
    observation_keys,
    expected_obs,
    expected_valid,
):
    config_data = {"observations": observation_filter}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, observation_keys
    )

    if expected_valid:
        assert config.valid, config.errors
        assert sorted(config.snapshot.observations) == sorted(expected_obs)
    else:
        assert not config.valid
        assert 1 == len(config.errors)
        expected_key_path = ("observations",)
        assert expected_key_path == config.errors[0].key_path[:-1]


@pytest.mark.parametrize(
    "observation_keys",
    [
        ("my_obs", "other_obs"),
        ("my_obs", "other_obs", "obs1") + tuple(i * "a" for i in range(20)),
        ("my_obs", "other_obs"),
    ],
)
def test_default_observations(observation_keys):
    config_data = {}
    config = misfit_preprocessor.config.MisfitPreprocessorConfig(
        config_data, observation_keys
    )

    assert config.valid, config.errors
    assert sorted(observation_keys) == sorted(config.snapshot.observations)



import logging
from copy import deepcopy

import numpy as np

from semeio.workflows.correlated_observations_scaling.exceptions import (
    EmptyDatasetException,
)


class DataMatrix(object):
    def __init__(self, input_data):
        """
        Takes input data in the form of a Pandas multi index dataframe with
        observations, standard deviation and simulated values. Assumes observations
        are prepended with _OBS and standard deviation with _STD.
        """
        self.data = input_data
        if input_data.shape[1] == 0:
            raise EmptyDatasetException("Empty dataset")

    def get_data_matrix(self):
        """
        Extracts data from a dataframe and returns a matrix
        """
        return self.data[~self.data.index.isin(["OBS", "STD"])].values

    def normalize_by_std(self):
        """
        Normalize the matrix in place.
        """
        self.data = self.get_normalized_by_std()

    def get_normalized_by_std(self):
        """
        Duplicates the behavior of obs_data_scale, and scales the simulated data
        by 1 / (observation standard deviation), per observation key, i.e. each
        simulation data point is scaled by its corresponding std deviation
        from observations and returns a copy.
        """
        output_data = deepcopy(self.data)
        data_matrix = self._get_data()
        std_vector = self.data.loc["STD"]
        output_data[~output_data.index.isin(["OBS", "STD"])] = data_matrix * (
            1.0 / std_vector
        )

        return output_data

    def get_scaling_factor(self, nr_components):
        """
        Calculates a observation scaling factor which is:
            sqrt(nr_obs / pc)
        where:
            nr_obs is the number of observations
            pc is the number of primary components from PCA analysis
                below a user threshold
        """
        data_matrix = self.get_data_matrix()
        nr_observations = data_matrix.shape[1]

        logging.info(
            (
                "Calculation scaling factor, nr of primary components: "
                "{:d}, number of observations: {:d}"
            ).format(nr_components, nr_observations)
        )
        return np.sqrt(nr_observations / float(nr_components))

    def _get_data(self):
        return self.data[~self.data.index.isin(["OBS", "STD"])]

    def get_nr_primary_components(self, threshold):
        """
        Takes a matrix, does PCA and calculates the cumulative variance ratio
        and returns an int which is the number of primary components where
        the cumulative variance is smaller than user set threshold. Notice the
        way of calculating number of primary components. This is done to
        replicate existing behavior:

        int num_significant  = 0;
        {
        double running_sigma2  = 0;
        for (int i=0; i < num_singular_values; i++) {
          if (running_sigma2 / total_sigma2 < truncation) {
             num_significant++;
             running_sigma2 += sig0[i] * sig0[i];
          } else
             break;
        }

        Also returns an array of singular values.
        """
        data_matrix = self.get_data_matrix()
        data_matrix = data_matrix - data_matrix.mean(axis=0)
        _, s, _ = np.linalg.svd(data_matrix.astype(np.float), full_matrices=False)
        variance_ratio = np.cumsum(s ** 2) / np.sum(s ** 2)
        return len([1 for i in variance_ratio[:-1] if i < threshold]) + 1, s



import numpy as np
import pandas as pd

import warnings
import logging


# Filenames created/modified by this script.
# Don't change unless you understand the consequences.
_TARGET_FILE_TXT = "DESIGN2PARAMS.OK"
_DESIGN_MATRIX_TXT = "designmatrix.txt"
_DESIGN_PARAMETERS_TXT = "designparameters.txt"
_PARAMETERS_TXT = "parameters.txt"

logger = logging.getLogger(__name__)


def run(
    realization,
    xlsfilename,
    designsheetname,
    defaultssheetname,
    parametersfilename,
    log_level,
):
    """
    Reads out all file content from different files and create dataframes
    """
    logger.setLevel(log_level)

    design_matrix_sheet = _read_excel(xlsfilename, designsheetname)
    _validate_design_matrix(design_matrix_sheet)

    default_df = _read_defaultssheet(xlsfilename, defaultssheetname)

    try:
        parameters = pd.read_csv(parametersfilename, delimiter=" ", header=None)
    except IOError:
        logger.info(
            "No {} exists, creating a new, empty one.".format(parametersfilename)
        )
        parameters = pd.DataFrame(columns=[0, 1])
    parameters.rename(columns={0: "keys", 1: "parameters"}, inplace=True)

    _complete_parameters_file(
        realization, parameters, parametersfilename, design_matrix_sheet, default_df
    )


def _complete_parameters_file(
    realization, parameters, parametersfilename, design_matrix_sheet, default_sheet
):
    """
    Pick key / values from choosen realization in design matrix
    Append those key / values if not present into parameters.txt
    Create file adding those key / values to designparameters.txt
    if default-sheet contains key / values not existing in
    design matrix nor parameters, append those as well to both
    parameters.txt and designparameters.txt
    Write designmatrix.txt as a csv repr of xls

    :raises: SystemExit if design matrix contains empty headers
    :raises: SystemExit design matrix contains empty cells
    :raises: SystemExit if provided realization not exists
    """
    # get header / vals for choosen realization
    try:
        realization_values = pd.DataFrame(design_matrix_sheet.iloc[realization, 1:])
    except IndexError:
        raise SystemExit(
            "Provided realization arg {} does not exist in design matrix".format(
                realization
            )
        )
    realization_values.reset_index(inplace=True)
    realization_values.rename(
        columns={"index": "keys", realization: "realization"}, inplace=True
    )

    # merge realization from design-matrix with parameters and defaults to one dataframe
    merged = pd.merge(
        parameters,
        pd.merge(realization_values, default_sheet, on="keys", how="outer"),
        on="keys",
        how="outer",
    )

    # add new column with a combined parameters / data
    # from the realization in design matrix
    merged["parameters_realization"] = merged["parameters"].combine_first(
        merged["realization"]
    )

    # add new column with a total combined of parameters_realization and defaults
    merged["combined"] = merged["parameters_realization"].combine_first(
        merged["defaults"]
    )

    # keys in the design matrix not present in parameters
    design_parameters = merged[
        merged["parameters"].isnull() & merged["realization"].notnull()
    ]
    logger.info(
        "\ndesign parameters used: \n%s", design_parameters[["keys", "combined"]]
    )

    # keys without value in parameters and design matrix
    defaults = merged[merged["parameters_realization"].isnull()]
    logger.info("\ndefaults used: \n%s", defaults[["keys", "combined"]])

    # write used design matrix to csv
    design_matrix_sheet.to_csv(
        _DESIGN_MATRIX_TXT,
        sep=" ",
        index=False,
        float_format="%.10f",
        na_rep="__VOID__",
    )

    # keys not present in parameters
    combined_parameters = merged[merged["parameters"].isnull()]

    # append new keys to parameters.txt
    combined_parameters.to_csv(
        parametersfilename,
        columns=["keys", "combined"],
        sep=" ",
        mode="a",
        header=False,
        index=False,
    )

    # write file with the new keys to designparameters.txt
    combined_parameters.to_csv(
        _DESIGN_PARAMETERS_TXT,
        columns=["keys", "combined"],
        sep=" ",
        header=False,
        index=False,
    )

    # if all ok - write the ok file
    with open(_TARGET_FILE_TXT, "w") as target_file:
        target_file.write("OK\n")


def _read_excel(file_name, sheet_name, header=0):
    """
    Make dataframe from excel file
    :return: Dataframe
    :raises: SystemExit if file not found
    :raises: SystemExit if file not loaded correctly
    """
    try:
        return pd.read_excel(file_name, sheet_name, header=header, dtype=str)
    except IOError:
        raise SystemExit("File {} not found".format(file_name))
    except Exception as err:
        raise SystemExit(
            (
                "File {} is probably not of correct type. Failed with exception '{}'"
            ).format(file_name, str(err))
        )


def _validate_design_matrix(design_matrix):
    """
    Validate user inputted design matrix
    :raises: SystemExit if design matrix contains empty headers
    :raises: SystemExit design matrix contains empty cells
    """
    # find column headers missing and raise exception if any
    unnamed = design_matrix.loc[:, design_matrix.columns.str.contains("^Unnamed")]
    column_indexes = [int(x.split(":")[1]) for x in unnamed.columns.values]
    if len(column_indexes) > 0:
        raise SystemExit(
            "Column headers not present in column {}".format(column_indexes)
        )

    # find empty cells and raise exception if any
    empties = [
        "Realization {}, key {}".format(i, j)
        for i, j in zip(*np.where(pd.isnull(design_matrix)))
    ]
    if len(empties) > 0:
        raise SystemExit("Design matrix contains empty cells {}".format(empties))


def _read_defaultssheet(xlsfilename, defaultssheetname):
    """
    Reads a XLSX file and tries to return a dataframe
    with defaults for design matrix parameters.

    A dataframe is always returned, possibly empty.
    Columns are always exactly named "keys" and "defaults"

    :raises: SystemExit if defaults sheet is non-empty but non-parsable
    """
    if defaultssheetname:
        default_df = _read_excel(xlsfilename, defaultssheetname, header=None)
        if default_df.empty:
            logger.info("Empty defaultssheet provided")
            default_df = pd.DataFrame(columns=[0, 1])
        if len(default_df.columns) < 2:
            raise SystemExit(
                "Defaults sheet must have exactly two columns, only one found"
            )
        if len(default_df.columns) > 2:
            warnings.warn(
                (
                    "DEPRECATION: You supplied more than two columns for "
                    "the default values. This is not supported and will stop "
                    "working in the future."
                ),
                UserWarning,
            )
            default_df = default_df[[0, 1]]  # Slicing columns
    else:
        logger.info("No defaultssheet provided, using empty dataframe")
        default_df = pd.DataFrame(columns=[0, 1])

    default_df.rename(columns={0: "keys", 1: "defaults"}, inplace=True)
    return default_df



import json
import os
import pytest
import shutil
import subprocess
import sys
from semeio.communication import SEMEIOSCRIPT_LOG_FILE


TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "test-data")
ERT_INSTALLED = shutil.which("ert") is not None


@pytest.mark.skipif(not ERT_INSTALLED, reason="ERT is not installed")
def test_semeio_script_integration(tmpdir):
    tmpdir.chdir()

    shutil.copytree(TEST_DATA_DIR, "test_data")
    os.chdir(os.path.join("test_data"))

    ert_env = {
        env_var: os.environ[env_var]
        for env_var in ("PATH", "LD_LIBRARY_PATH")
        if env_var in os.environ
    }
    ert_env["PYTHONPATH"] = os.pathsep.join(map(os.path.realpath, sys.path))

    subprocess.check_call(
        ("ert", "workflow", "TEST_WORKFLOW", "config.ert"),
        env=ert_env,
    )

    # Assert that data was published correctly
    with open("reports/config/TestWorkflowJob/test_data.json") as f:
        reported_data = json.load(f)
    assert [list(range(10))] == reported_data

    # Assert that logs were forwarded correctly
    log_file = os.path.join("reports/config/TestWorkflowJob/", SEMEIOSCRIPT_LOG_FILE)
    with open(log_file) as f:
        log = f.readlines()
    assert len(log) == 1
    expected_log_msg = (
        "I finished without any problems - hence I'm not a failure after all!"
    )
    assert expected_log_msg in log[0]



import logging
import os
from distutils.dir_util import copy_tree

import pytest

from semeio.jobs.design_kw import design_kw


@pytest.fixture
def input_data(tmpdir):
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
    copy_tree(data_dir, tmpdir.strpath)

    cwd = os.getcwd()
    tmpdir.chdir()

    yield

    os.chdir(cwd)


def test_extract_key_value_ok():
    data = [
        "key1 14",
        "key2 24",
        "key3 34",
    ]
    assert design_kw.extract_key_value(data) == {
        "key1": "14",
        "key2": "24",
        "key3": "34",
    }


def test_extract_value_with_space():
    data = [
        'key4 "foo bar com"',
        'key5 "conserve   spaces"',
    ]
    assert design_kw.extract_key_value(data) == {
        "key4": "foo bar com",
        "key5": "conserve   spaces",
    }


def test_extract_key_with_space():
    data = ['"key x" "value with space"']
    assert design_kw.extract_key_value(data) == {"key x": "value with space"}


def test_extract_too_many_values():
    """Test that user friendly error message is displayed
    when an error occurs"""

    data = ["SENTENCE this is a string"]
    with pytest.raises(ValueError, match="SENTENCE"):
        design_kw.extract_key_value(data)
    with pytest.raises(ValueError, match="Too many values found"):
        design_kw.extract_key_value(data)


def test_unsupported_quoting():
    data = ['Unclosed " quote']
    with pytest.raises(ValueError, match="No closing quotation"):
        design_kw.extract_key_value(data)


def test_extract_missing_value():
    data = ["key1"]
    with pytest.raises(ValueError, match="No value found"):
        design_kw.extract_key_value(data)


def test_extract_key_value_dup_key():
    data = ["key1 14", "key2 24", "key1 34"]
    with pytest.raises(ValueError, match="multiple times"):
        design_kw.extract_key_value(data)


@pytest.mark.parametrize(
    "paramsdict, expected",
    [
        ({"foo:bar": "value"}, {"bar": "value"}),
        ({"foo:bar": "value", "foo:com": "value2"}, {"bar": "value", "com": "value2"}),
        ({}, {}),
    ],
)
def test_rm_genkw_prefix(paramsdict, expected):
    assert design_kw.rm_genkw_prefix(paramsdict) == expected


@pytest.mark.parametrize(
    "paramsdict, warntext",
    [
        (
            {"foo:bar": "value1", "com:bar": "value2"},
            "Key(s) ['bar'] can only be used with prefix",
        ),
        (
            {"foo:bar": "value1", "bar": "value2"},
            "Key(s) ['bar'] can only be used with prefix",
        ),
    ],
)
def test_rm_genkw_prefix_warnings(paramsdict, warntext, caplog):
    assert design_kw.rm_genkw_prefix(paramsdict) == {}
    assert warntext in caplog.text


@pytest.mark.parametrize(
    "paramsdict, ignores, expected",
    [
        (
            {"LOG10_foo:bar": "value1", "com:bar": "value2"},
            ["LOG10"],
            {"bar": "value2"},
        ),
        ({"LOG10_foo:bar": "value1", "com:bar": "value2"}, ["LOG"], {"bar": "value2"}),
        ({"LOG10PARAM": "value"}, ["LOG10"], {"LOG10PARAM": "value"}),
        ({"foo": "value"}, [""], {"foo": "value"}),
        ({"foo:bar": "value"}, [""], {"bar": "value"}),
    ],
)
def test_rm_genkw_prefix_ignore(paramsdict, ignores, expected):
    assert design_kw.rm_genkw_prefix(paramsdict, ignores) == expected


@pytest.mark.parametrize(
    "paramlines, template, result",
    [
        ("foo:bar 12", "<foo:bar>", "12"),
        ("foo:bar 12", "<bar>", "12"),
        ("foo:bar 12\ncom:bar 13", "<bar>", "<bar>"),  # Yields logged error
        ("LOG10_FOO:foo 1\nFOO:foo 10", "<foo>", "10"),
        ("LOG30_FOO:foo 0.3\nFOO:foo 10", "<foo>", "<foo>"),  # Error logged
        ("foo:bar 12\nbar 12", "<bar>", "12"),
        ("foo:bar 13\nbar 12", "<foo:bar>", "13"),
        ("foo:bar 13\ncom:bar 14", "<bar>", "<bar>"),
        ("foo:bar 13\ncom:bar 14", "<foo:bar>", "13"),
        ("foo:bar 13\ncom:bar 14", "<com:bar>", "14"),
    ],
)
def test_genkw_prefix_handling(paramlines, template, result, tmpdir):
    tmpdir.chdir()
    with open("parameters.txt", "w") as file_h:
        file_h.write(paramlines)
    with open("template.tmpl", "w") as file_h:
        file_h.write(template)
    design_kw.run("template.tmpl", "result.txt", logging.DEBUG)
    with open("result.txt", "r") as file_h:
        resulttxt = "\n".join(file_h.readlines())
    assert resulttxt == result


def test_is_comment():
    line = "# this is a std pattern comment"
    assert design_kw.is_comment(line)

    line = "-- this is a ecl comment"
    assert design_kw.is_comment(line)

    line = "  ECLIPSE_INIT_DATE: [2000-01-01]"
    assert not design_kw.is_comment(line)


def test_unmatched_templates():
    line = "  ECLIPSE_INIT_DATE: [2000-01-01]"
    assert not design_kw.unmatched_templates(line)

    line = "  ECLIPSE_INIT_DATE: <DATE>"
    assert design_kw.unmatched_templates(line) == ["<DATE>"]


def test_is_perl():
    file_name = "filename.yml"
    template = ["perl", "  ECLIPSE_INIT_DATE: <DATE>"]
    assert design_kw.is_perl(file_name, template)

    file_name = "filename.pl"
    template = ["perl", "  ECLIPSE_INIT_DATE: <DATE>"]
    assert design_kw.is_perl(file_name, template)

    file_name = "filename.pl"
    template = ["  ECLIPSE_INIT_DATE: <DATE>"]
    assert design_kw.is_perl(file_name, template)

    file_name = "filename.yml"
    template = ["  ECLIPSE_INIT_DATE: <DATE>"]
    assert not design_kw.is_perl(file_name, template)


@pytest.mark.parametrize(
    "filenames",
    (
        ["template.yml.reference", "template.yml.tmpl", "template.yml"],
        ["schfile.inc.reference", "schfile.inc.tmpl", "schfile.inc"],
    ),
)
def test_run(input_data, filenames):
    reference_filename = filenames[0]
    template_filename = filenames[1]
    result_filename = filenames[2]

    design_kw.run(template_filename, result_filename, log_level=logging.DEBUG)

    with open(result_filename, "r") as result_file:
        result = result_file.read()

    with open(reference_filename, "r") as reference_file:
        reference = reference_file.read()

    with open(design_kw._STATUS_FILE_NAME, "r") as status_file:
        status = status_file.read()

    assert result == reference
    assert status == "DESIGN_KW OK\n"


def test_run_unmatched(input_data):
    template_filename = "template_unmatched.yml.tmpl"
    result_filename = "template.yml"
    design_kw.run(
        template_file_name=template_filename,
        result_file_name=result_filename,
        log_level=logging.DEBUG,
        parameters_file_name="parameters.txt",
    )

    assert not os.path.isfile(design_kw._STATUS_FILE_NAME)


def test_run_duplicate_keys(input_data):
    template_filename = "template.yml.tmpl"
    result_filename = "template.yml"
    with pytest.raises(ValueError, match="multiple"):
        design_kw.run(
            template_file_name=template_filename,
            result_file_name=result_filename,
            log_level=logging.DEBUG,
            parameters_file_name="parameters_w_duplicates.txt",
        )

    assert not os.path.isfile(design_kw._STATUS_FILE_NAME)



import numpy as np
import pandas as pd
import pytest
import random
from semeio.workflows import misfit_preprocessor

from unittest.mock import Mock


class MockedMeasuredData(object):
    def __init__(self, observations, responses):
        self._data = self._build_data(observations, responses)

    def _build_data(self, observations, responses):
        def assert_equal_keys(a, b):
            assert sorted(a.keys()) == sorted(b.keys())

        columns = []
        ens_size = None
        assert_equal_keys(observations, responses)
        for key in observations.keys():
            assert_equal_keys(observations[key], responses[key])
            for sub_key, sim in responses[key].items():
                if ens_size is None:
                    ens_size = len(sim)
                assert ens_size == len(sim)
                columns.append((key,) + tuple(sub_key))

        index = ("OBS", "STD") + tuple(range(ens_size))

        raw_data = np.array(
            [
                np.array([observations[col[0]][col[1:]][0] for col in columns]),
                np.array([observations[col[0]][col[1:]][1] for col in columns]),
            ]
        )
        simulated_raw_data = np.array([responses[col[0]][col[1:]] for col in columns]).T
        raw_data = np.concatenate((raw_data, simulated_raw_data))

        data = pd.DataFrame(
            raw_data,
            index=index,
            columns=pd.MultiIndex.from_tuples(
                columns, names=[None, "key_index", "data_index"]
            ),
        )

        return data

    @property
    def data(self):
        return self._data

    def get_simulated_data(self):
        return self._data.drop(index=["OBS", "STD"])

    def remove_failed_realizations(self):
        pass

    def remove_inactive_observations(self):
        pass

    def filter_ensemble_std(self, _):
        pass


def assert_homogen_clusters(config):
    for sub_config in config:
        assert 1 == len(sub_config["CALCULATE_KEYS"]["keys"])


def generate_simulated_responses(
    forward_polynomials,
    parameter_distribution,
    poly_states,
    ensemble_size,
):
    simulated = {}
    for poly_idx, (poly_fm, states) in enumerate(zip(forward_polynomials, poly_states)):
        new_parameters = np.random.uniform(0, 10, 3 * ensemble_size)
        new_parameters.resize(ensemble_size, 3)
        simulated["poly_{}".format(poly_idx)] = {
            (state, state): sum((new_parameters * np.array((state ** 2, state, 1))).T)
            for state in states
        }

    return simulated


def generate_observations(
    forward_polynomials,
    parameter_distribution,
    poly_states,
):
    true_parameters = parameter_distribution()

    return {
        "poly_{}".format(poly_idx): {
            (state, state): (
                poly_fm(
                    a=true_parameters[poly_idx]["a"],
                    b=true_parameters[poly_idx]["b"],
                    c=true_parameters[poly_idx]["c"],
                    x=state,
                ),
                1,
            )
            for state in states
        }
        for poly_idx, (poly_fm, states) in enumerate(
            zip(forward_polynomials, poly_states)
        )
    }


def generate_measurements(num_polynomials, poly_states=None, ensemble_size=10000):
    if poly_states is None:
        poly_states = [range(3) for _ in range(num_polynomials)]

    forward_polynomials = [
        lambda a, b, c, x: a * x ** 2 + b * x + c for _ in range(num_polynomials)
    ]

    def parameter_distribution():
        return [
            {
                "a": random.uniform(0, 10),
                "b": random.uniform(0, 10),
                "c": random.uniform(0, 10),
            }
            for _ in enumerate(forward_polynomials)
        ]

    observations = generate_observations(
        forward_polynomials,
        parameter_distribution,
        poly_states,
    )
    simulated = generate_simulated_responses(
        forward_polynomials,
        parameter_distribution,
        poly_states,
        ensemble_size,
    )
    return observations, simulated


@pytest.mark.parametrize(
    "num_polynomials",
    tuple(range(1, 5)) + (20, 100),
)
def test_misfit_preprocessor_n_polynomials(num_polynomials):
    state_size = 3
    poly_states = [range(1, state_size + 1) for _ in range(num_polynomials)]

    observations, simulated = generate_measurements(
        num_polynomials,
        poly_states=poly_states,
        ensemble_size=10000,
    )
    measured_data = MockedMeasuredData(observations, simulated)

    config = {}
    reporter_mock = Mock()
    configs = misfit_preprocessor.run(config, measured_data, reporter_mock)
    assert_homogen_clusters(configs)
    assert num_polynomials == len(configs), configs


@pytest.mark.parametrize(
    "state_size",
    [5 * [30], [5, 5, 5, 5, 100]],
)
def test_misfit_preprocessor_state_size(state_size):
    num_polynomials = 5
    poly_states = [range(1, size + 1) for size in state_size]

    observations, simulated = generate_measurements(
        num_polynomials,
        poly_states=poly_states,
        ensemble_size=30000,
    )
    measured_data = MockedMeasuredData(observations, simulated)

    config = {}
    reporter_mock = Mock()
    configs = misfit_preprocessor.run(config, measured_data, reporter_mock)
    assert_homogen_clusters(configs)
    assert num_polynomials == len(configs), configs


@pytest.mark.parametrize(
    "state_size",
    [5 * [30], [5, 5, 5, 5, 100], [5, 5, 5, 5, 100], [2, 1000]],
)
def test_misfit_preprocessor_state_uneven_size(state_size):
    num_polynomials = len(state_size)
    poly_states = [range(1, size + 1) for size in state_size]

    observations, simulated = generate_measurements(
        num_polynomials,
        poly_states=poly_states,
        ensemble_size=30000,
    )
    measured_data = MockedMeasuredData(observations, simulated)

    config = {
        "clustering": {
            "spearman_correlation": {
                "fcluster": {"t": num_polynomials + 1, "criterion": "maxclust"}
            }
        }
    }
    reporter_mock = Mock()
    configs = misfit_preprocessor.run(config, measured_data, reporter_mock)
    assert num_polynomials == len(configs), configs
    assert_homogen_clusters(configs)


def test_misfit_preprocessor_configuration_errors():
    observations, simulated = generate_measurements(1)
    measured_data = MockedMeasuredData(observations, simulated)

    config = {
        "unknown_key": [],
        "clustering": {"spearman_correlation": {"fcluster": {"threshold": 1.0}}},
    }
    reporter_mock = Mock()
    with pytest.raises(misfit_preprocessor.ValidationError) as ve:
        misfit_preprocessor.run(config, measured_data, reporter_mock)

    expected_err_msg = (
        "Invalid configuration of misfit preprocessor\n"
        "  - Unknown key: unknown_key (root level)\n"
        "  - Unknown key: threshold (clustering.spearman_correlation.fcluster)\n"
    )
    assert expected_err_msg == str(ve.value)



# -*- coding: utf-8 -*-
import pytest

from semeio.workflows.spearman_correlation_job import job as spearman


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        ([(1, "KEY_1", 1)], {1: {"KEY_1": [1]}}),
        ([(1, "KEY_1", 1), (1, "KEY_1", 2)], {1: {"KEY_1": [1, 2]}}),
        ([(1, "KEY_1", 1), (2, "KEY_2", 2)], {1: {"KEY_1": [1]}, 2: {"KEY_2": [2]}}),
    ],
)
def test_make_clusters(test_input, expected_result):
    result = spearman._cluster_data(test_input)
    assert result == expected_result


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        ({}, {}),
        ({1: {"KEY_1": [1]}}, {}),
        ({1: {"KEY_1": [1, 2]}}, {0: {"KEY_1": [1, 2]}}),
        (
            {1: {"KEY_1": [1], "KEY_2": [2]}, 2: {"KEY_2": [2]}},
            {0: {"KEY_1": [1], "KEY_2": [2]}},
        ),
    ],
)
def test_single_obs_clusters(test_input, expected_result):
    result = spearman._remove_singular_obs(test_input)
    assert result == expected_result, "bad removal of singular obs clusters"


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (
            {1: {"KEY_1": [1]}},
            [{"CALCULATE_KEYS": {"keys": [{"key": "KEY_1", "index": [1]}]}}],
        )
    ],
)
def test_config_creation(test_input, expected_result):
    result = spearman._config_creation(test_input)
    assert result == expected_result



import argparse
import os

from semeio.jobs.rft.utility import strip_comments


class ZoneMap:
    """A zonemap is a map from simulation grid
    k layers to a list of strings representing the
    possible zones (in a stratigraphical hiearchy of zones)
    that the k-layers belongs to.


    Args
        zones_at_k_value (dict): A dictionary from each
            k layer (as integer key larger than 0) to a list
            of strings with zone names.
    """

    def __init__(self, zones_at_k_value=None):
        self._zones_at_k_value = zones_at_k_value or {}
        self._k_values_at_zone = {}

        # Construct a reverse dictionary for the reverse
        # lookup _k_values_at_zone
        for k_value, zone_names in self._zones_at_k_value.items():
            for zone_name in zone_names:
                if zone_name not in self._k_values_at_zone:
                    self._k_values_at_zone[zone_name] = []

                self._k_values_at_zone[zone_name].append(k_value)

    @classmethod
    def load_and_parse_zonemap_file(cls, filename):
        # The job description files used in ERT does not allow
        # for optional arguments. If the user has not specified
        # a value in their configuration, the job description
        # uses the default value ZONEMAP_NOT_PROVIDED
        if filename == "ZONEMAP_NOT_PROVIDED":
            return None

        if not os.path.isfile(filename):
            raise argparse.ArgumentTypeError(
                "ZoneMap file {filename} not found!".format(filename=filename)
            )

        zones_at_k_value = {}

        with open(filename, "r") as f:
            zonemap_lines = f.readlines()

        zonemap_lines = [
            (strip_comments(l), i + 1) for i, l in enumerate(zonemap_lines)
        ]
        basic_err_msg = (
            "Line {line_number} in ZoneMap file {filename} "
            "not on proper format: 'k zonename <zonename> ...'. "
        )
        for line, line_number in zonemap_lines:
            zonemap_line = line.split()

            if not zonemap_line:
                continue

            if len(zonemap_line) < 2:
                raise argparse.ArgumentTypeError(
                    basic_err_msg.format(line_number=line_number, filename=filename)
                    + "Number of zonenames must be 1 or more",
                )
            try:
                raw_k = int(zonemap_line[0])
            except ValueError:
                raise argparse.ArgumentTypeError(
                    basic_err_msg.format(line_number=line_number, filename=filename)
                    + "k must be integer, was {k}".format(k=zonemap_line[0])
                )
            if raw_k == 0:
                raise argparse.ArgumentTypeError(
                    basic_err_msg.format(line_number=line_number, filename=filename)
                    + "k values cannot be 0, must start at 1. "
                )

            k_value = raw_k - 1
            zones = [zone.strip() for zone in zonemap_line[1:]]

            zones_at_k_value[k_value] = zones

        return cls(zones_at_k_value)

    def __contains__(self, item):
        if isinstance(item, int):
            return item in self._zones_at_k_value
        elif isinstance(item, str):
            return item in self._k_values_at_zone
        return False

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._zones_at_k_value[item]
        elif isinstance(item, str):
            return self._k_values_at_zone[item]
        raise KeyError("{item} is neither a k value nor a zone".format(item=item))

    def has_relationship(self, zone, k):
        return k in self._k_values_at_zone.get(zone, [])



def has_keys(main_list, sub_list, error_msg):
    """
    Checks that all sub_list are present and returns a list of error messages
    """
    return [error_msg.format(key) for key in sub_list if key not in main_list]


def is_subset(main_list, sub_list):
    """
    Checks if all the keys in sub_list are present in main_list and returns list of
    error messages
    """
    error_msg = "Update key: {} missing from calculate keys: {}"
    missing_keys = set(sub_list).difference(set(main_list))
    return [error_msg.format(missing_key, main_list) for missing_key in missing_keys]



#!/usr/bin/env python
"""Forward model for connecting ERT with the Pyscal command line client"""
import logging
import sys
import os
import argparse

from semeio.jobs.design_kw.design_kw import extract_key_value, rm_genkw_prefix

from pyscal import pyscalcli

_logger = logging.getLogger("FM_PYSCAL")

# The string used here must match what is used as the DEFAULT
# parameter in semeio/jobs/config_jobs/PYSCAL. It is not used elsewhere.
MAGIC_NONE = "__NONE__"

# These key-values are added to the dictionary parsed from
# parameters.txt, and allow supplying magic interpolation values
# directly from the ert config file.
MAGIC_CASES = {
    "__BASE__": 0,
    "__LOW__": -1,
    "__PESS__": -1,
    "__PESSIMISTIC__": -1,
    "__HIGH__": 1,
    "__OPT__": 1,
    "__OPTIMISTIC__": 1,
}


def main_entry_point(args=None):
    """This mimics the pyscal command line client, but differs because
    all arguments are required (but can be defaulted using MAGIC_NONE)"""
    parser = _get_args_parser()
    options = parser.parse_args(args)
    run(
        options.relperm_parameters_file,
        options.output_filename,
        options.sheet_name,
        options.int_param_wo_name,
        options.int_param_go_name,
        options.slgof,
        options.family,
    )


description = (
    "ERT forward model wrapping around the pyscal command line client. "
    "In the forward model context, this gives access to interpolation "
    "parameters in parameters.txt which the command line client is "
    "not aware of. For other uses, head to the pyscal client from "
    "the pyscal package."
)

category = "modeling.reservoir"


def _get_args_parser():
    """Construct an argparse parser for fm_pyscal"""

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "relperm_parameters_file",
        type=str,
        help=(
            "CSV or XLSX filename with relperm parameters. "
            "See pyscal documentation for table format."
        ),
    )
    parser.add_argument(
        "output_filename",
        type=str,
        help=("Location of Eclipse relperm include file to be written."),
        default="relperm.inc",
    )
    parser.add_argument(
        "sheet_name",
        type=str,
        help="XLSX sheetname to use. Will use the first sheet if not specified.",
        default=MAGIC_NONE,
    )
    parser.add_argument(
        "int_param_wo_name",
        type=str,
        help=(
            "Interpolation parameter name to be parsed from parameters.txt "
            "for WaterOil parameter if SCAL recommendation is given. "
            "You may also use the the mnemonics __OPT__, __BASE__ and __PESS__. "
            "The values in parameters.txt must be in the interval [-1,1]. "
            "For parameters generated by GEN_KW, do not include the namespace "
            "in front of the colon in the parameter name."
        ),
        default=MAGIC_NONE,
    )
    parser.add_argument(
        "int_param_go_name",
        type=str,
        help=(
            "Ditto for GasOil. If not supplied, the WaterOil parameter will be used."
        ),
        default=MAGIC_NONE,
    )
    parser.add_argument(
        "slgof",
        type=str,
        help="Set to slgof if SLGOF is wanted in place of SGOF. Case insensitive.",
        default="sgof",
    )
    parser.add_argument(
        "family",
        type=int,
        help=(
            "Family (i) (SWOF + SGOF) or family (ii) (SWFN + SOF3 + SGFN) "
            "for Eclipse keywords. Supply integer 1 or 2. Default family (i), 1."
        ),
        default=1,
    )
    return parser


def run(
    relperm_parameters_file,
    output_filename,
    sheet_name,  # string, use 0 or __NONE__ when irrelevant or default
    int_param_wo_name,  # string or __NONE__
    int_param_go_name,  # string or __NONE__
    slgof,  # sgof or slgof, default sgof
    family,  # int: 1 or 2, default 1
    parameters_file_name="parameters.txt",
):
    """This function is a wrapper around the Pyscal command
    line tool. The command line tool is designed around argparse
    and this function wraps around that design.

    In contrast with the command line tool where interpolation
    parameters are explicit, they are implicit here, and gathered
    from parameters.txt"""
    if not os.path.exists(relperm_parameters_file):
        _logger.error("%s does not exist", relperm_parameters_file)
        sys.exit(1)

    # Determine which interpolation scenario the user has requested:
    if int_param_wo_name != MAGIC_NONE and int_param_go_name != MAGIC_NONE:
        # Separate interpolation parameter for WaterOil and GasOil
        do_interpolation = True
    elif int_param_wo_name != MAGIC_NONE and int_param_go_name == MAGIC_NONE:
        # In this scenario, the WaterOil interpolation parameter
        # should be used for GasOil as well.
        do_interpolation = True
        int_param_go_name = int_param_wo_name
    elif int_param_wo_name == MAGIC_NONE and int_param_go_name == MAGIC_NONE:
        do_interpolation = False
    else:
        # Something is wrong if we end here
        _logger.error("WaterOil interpolation parameter missing")
        sys.exit(1)

    slgof = slgof.lower()
    if slgof not in ["sgof", "slgof"]:
        _logger.error("Only supports sgof or slgof")
        sys.exit(1)

    if family not in [1, 2]:
        _logger.error("Family must be either 1 or 2")
        sys.exit(1)

    if do_interpolation:
        (int_param_wo, int_param_go) = _get_interpolation_values(
            int_param_wo_name, int_param_go_name, parameters_file_name
        )
    else:
        int_param_wo = None
        int_param_go = None

    if sheet_name in ["0", MAGIC_NONE]:
        # Limitation: If the user names an xls sheet with the string "0"
        # and it is not the first sheet, it will not be accessible with this
        # forward model
        sheet_name = None

    try:
        pyscalcli.pyscal_main(
            parametertable=relperm_parameters_file,
            verbose=True,
            output=output_filename,
            sheet_name=sheet_name,
            int_param_wo=int_param_wo,
            int_param_go=int_param_go,
            slgof=slgof == "slgof",
            family2=family == 2,
        )
    except ValueError as e_msg:
        _logger.error(str(e_msg))
        raise e_msg


def _get_interpolation_values(
    int_param_wo_name, int_param_go_name, parameters_file_name="parameters.txt"
):
    """
    Given parameter names, obtain values to interpolate through from parameters.txt

    If only WaterOil is supplied, the GasOil interpolation value will
    be copied from the WaterOil value.

    Args:
        int_param_wo_name (string): parameter name (no genkw_prefix) for WaterOil
        int_param_go_name (string): parameter name (no genkw_prefix) for GasOil
        parameters_file_name (string): Text file to look for parameters in.

    Returns:
        tuple with two values, one for WaterOil and one for GasOil
    """
    # Read all key-value pairs from parameters.txt
    if not os.path.exists(parameters_file_name):
        _logger.error("%s does not exists", parameters_file_name)
        raise IOError
    with open(parameters_file_name) as parameters_file:
        parameters = parameters_file.readlines()
    parameter_dict = extract_key_value(parameters)
    parameter_dict.update(rm_genkw_prefix(parameter_dict))

    # Add some magic parameters:
    parameter_dict.update(MAGIC_CASES)
    if int_param_wo_name not in parameter_dict:
        _logger.error(
            "Requested parameter name %s not found in %s",
            int_param_wo_name,
            parameters_file_name,
        )
        raise ValueError("Parameter name not found")
    int_param_wo = float(parameter_dict[int_param_wo_name])
    _logger.info(
        "Collected %s from parameter.txt with value %f",
        int_param_wo_name,
        int_param_wo,
    )
    int_param_go = int_param_wo

    if int_param_go_name not in parameter_dict:
        _logger.error(
            "Requested parameter name %s not found in %s",
            int_param_go_name,
            parameters_file_name,
        )
        raise ValueError("Parameter name not found")
    int_param_go = float(parameter_dict[int_param_go_name])
    if int_param_go_name != int_param_wo_name:
        _logger.info(
            "Collected %s from parameter.txt with value %f",
            int_param_go_name,
            int_param_go,
        )
    return (int_param_wo, int_param_go)


if __name__ == "__main__":
    main_entry_point()



import pytest

from semeio.workflows.correlated_observations_scaling import validator


@pytest.mark.parametrize(
    "test_input,expected_result",
    [
        (["POLY_OBS"], 0),
        (["NOT_A_KEY"], 1),
        (["POLY_OBS", "NOT_A_KEY"], 1),
        (["POLY_NOT", "NOT_A_KEY"], 2),
    ],
)
def test_has_keys(test_input, expected_result):
    obs = ["POLY_OBS"]
    msg = "fail_message"
    assert len(validator.has_keys(obs, test_input, msg)) == expected_result


@pytest.mark.parametrize(
    "input_list,result_list", [(["a"], []), (["a", "c"], []), (["a", "b", "c"], [])]
)
def test_is_subset_valid(input_list, result_list):
    example_list = ["a", "b", "c"]

    assert validator.is_subset(example_list, input_list) == result_list


@pytest.mark.parametrize(
    "input_list,list_length", [(["d"], 1), (["d", "e"], 2), (["a", "b", "d"], 1)]
)
def test_is_subset_invalid(input_list, list_length):
    example_list = ["a", "b", "c"]

    assert len(validator.is_subset(example_list, input_list)) == list_length



import argparse
import logging
import os
import sys

logger = logging.getLogger(__name__)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
std_out_handler = logging.StreamHandler(sys.stdout)
std_out_handler.setFormatter(formatter)
std_out_handler.setLevel(logging.INFO)
logger.addHandler(std_out_handler)

std_err_handler = logging.StreamHandler(sys.stdout)
std_err_handler.setFormatter(formatter)
std_err_handler.setLevel(logging.WARNING)
logger.addHandler(std_err_handler)


def valid_file(arg):
    if os.path.isfile(arg):
        return arg
    raise argparse.ArgumentTypeError("{} is not a existing file!".format(arg))
